"""
Utility functions for AP sensing module.
"""

import dascore as dc
from dascore.core import get_coord, get_coord_manager
from dascore.utils.misc import _maybe_unpack, unbyte


def _get_version_string(resource):
    """Determine if the file is AP sensing and what version."""
    attrs = {"AppVersion", "FileVersion", "OpticalChannelNumber"}
    groups = {"DAQ", "Interrogator", "Metadata", "ProcessingServer"}
    attrs_names = set(resource.attrs)
    group_names = set(resource)
    # This file doesn't have expected attrs and groups.
    if not attrs.issubset(attrs_names) and not groups.issubset(group_names):
        return False
    file_version = resource.attrs["FileVersion"]
    return str(_maybe_unpack(file_version))


def _get_time_coord(resource, shape):
    """Create the time coordinate."""
    # Load needed data from file and perform sanity checks.
    sr = _maybe_unpack(resource["ProcessingServer"]["DataRate"])
    start_time_str = unbyte(_maybe_unpack(resource["Metadata"]["Timestamp"]))
    time_zone = unbyte(_maybe_unpack(resource["Metadata"]["Timezone"]))
    assert time_zone in {"UTC", "UTC+0"}, "can't handle non-UTC timezones."
    trace_count = _maybe_unpack(resource["Metadata"]["TraceCount"])
    assert trace_count == shape[0], "trace count doesn't match dim shape"
    # Create coord
    start = dc.to_datetime64(start_time_str)
    step = dc.to_timedelta64(1 / sr)
    stop = start + trace_count * step
    coord = get_coord(start=start, stop=stop, step=step, units="s")
    return coord


def _get_distance_coord(resource, data_shape):
    """Get distance coordinate from AP sensing data."""
    start = resource["DAQ"]["PositionStart"]
    stop = resource["DAQ"]["PositionEnd"]
    start_unit = unbyte(_maybe_unpack(start.attrs.get("Unit", "m")))
    stop_unit = unbyte(_maybe_unpack(stop.attrs.get("Unit", "m")))
    assert start_unit == stop_unit

    # Oddly, SpatialSampling in ProcessServer group doesn't quite work to get
    # the right shape, so we either have to recalculate the spatial step, or
    # slightly change the end of the cable. Here we recalculate the sampling
    # rate, which is off from the stated rate by ~0.08% in ap_sensing_1 file.
    # step = _maybe_unpack(resource['ProcessingServer']['SpatialSampling'])
    dist_length = data_shape[1]
    x_start = _maybe_unpack(start)
    # x_end_calc = x_start + step * dist_length
    x_end = _maybe_unpack(stop)
    step_calc = (x_end - x_start) / dist_length
    coord = get_coord(start=x_start, stop=x_end, step=step_calc, units=start_unit)
    coord = coord.change_length(dist_length)
    return coord


def _get_coords(resource):
    """Get coordinates of AP_sensing file."""
    data_shape = resource["DAS"].shape
    # first get time
    time_coord = _get_time_coord(resource, data_shape)
    distance_coord = _get_distance_coord(resource, data_shape)
    cm = get_coord_manager(
        {"time": time_coord, "distance": distance_coord},
        dims=("time", "distance"),
    )
    return cm


def _get_attrs_dict(resource):
    """Get attributes."""
    daq = resource["DAQ"]
    pserver = resource["ProcessingServer"]
    out = dict(
        coords=_get_coords(resource),
        data_category="DAS",
        instrumet_id=unbyte(_maybe_unpack(daq["SerialNumber"])),
        gauge_length=_maybe_unpack(pserver["GaugeLength"]),
        radians_to_nano_strain=_maybe_unpack(pserver["RadiansToNanoStrain"]),
    )
    return out


def _get_patch(resource, time=None, distance=None, attr_cls=dc.PatchAttrs):
    """Get a patch from ap_sensing file."""
    attrs = _get_attrs_dict(resource)
    coords = attrs["coords"]
    data = resource["DAS"]
    if time is not None or distance is not None:
        coords, data = coords.select(array=data, time=time, distance=distance)
        attrs["coords"] = coords
    attrs = attr_cls.model_validate(attrs)
    return dc.Patch(data=data[:], coords=coords, attrs=attrs)



"""
Febus specific tests.
"""

import h5py
import numpy as np
import pytest

from dascore.io.febus import Febus2
from dascore.io.febus.utils import _flatten_febus_info
from dascore.utils.downloader import fetch
from dascore.utils.time import to_float


class TestFebus:
    """Special test cases for febus."""

    @pytest.fixture(scope="class")
    def febus_path(self):
        """Return the path to a test febus file."""
        return fetch("febus_1.h5")

    def test_time_coords_consistent_with_metadata(self, febus_path):
        """
        Ensure the time coords returned have the same length as
        metadata indicates.
        """
        patch = Febus2().read(febus_path)[0]
        coords = patch.coords
        time = coords.get_coord("time")
        time_span = to_float((time.max() - time.min()) + time.step)

        with h5py.File(febus_path, "r") as f:
            feb = _flatten_febus_info(f)[0]
            # First check total time extent
            n_blocks = feb.zone[feb.data_name].shape[0]
            block_time = 1 / (feb.zone.attrs["BlockRate"] / 1_000)
            expected_time_span = block_time * n_blocks
            assert np.isclose(expected_time_span, time_span)
            # Then check absolute time
            time_offset = feb.zone.attrs["Origin"][1] / 1_000
            time_start = feb.source["time"][0] + time_offset
            assert np.isclose(to_float(time.min()), time_start)



"""Tests for reading/writing pickles."""

from __future__ import annotations

import pickle
from io import BytesIO

import pytest

import dascore as dc
from dascore.io.pickle.core import PickleIO


@pytest.fixture(scope="session")
def pickle_patch_path(tmp_path_factory, random_patch):
    """Pickle a patch and return the path."""
    path = tmp_path_factory.mktemp("pickle_test") / "test.pkl"
    random_patch.io.write(path, "pickle")
    return path


class TestGetFormat:
    """Test detecting pickle format."""

    def test_detect_file(self, pickle_patch_path):
        """Simple test on output of is pickle."""
        parser = PickleIO()
        out = parser.get_format(pickle_patch_path)
        assert out
        assert out[0] == "PICKLE"

    def test_not_pickle(self, generic_hdf5):
        """Ensure non-pickle file returns false."""
        parser = PickleIO()
        assert not parser.get_format(generic_hdf5)

    def test_read_pickle(self, pickle_patch_path, random_patch):
        """Ensure a pickle file can be read."""
        out = dc.read(pickle_patch_path)
        assert isinstance(out, dc.BaseSpool)
        assert len(out) == 1
        assert isinstance(out[0], dc.Patch)
        assert random_patch == out[0]

    def test_file_not_there(self):
        """Get format should return false if the file doesn't exist."""
        parser = PickleIO()
        assert not parser.get_format("surely_not_a_file_that_exists")

    def test_get_format_binary_file_too_small(self):
        """Ensure a file which is too small returns false."""
        bio = BytesIO()
        bio.write(b"one")
        fio = PickleIO()
        bio.seek(0)
        assert not fio.get_format(bio)

    def test_has_dascore_not_pickle(self):
        """Test a buffer which has dascore but isnt a DC pickle."""
        bio = BytesIO()
        bio.write(b"dascore.core Spool")
        fio = PickleIO()
        bio.seek(0)
        assert not fio.get_format(bio)

    def test_bytes_io_valid_pickle(self, random_patch):
        """Test a valid pickle in bytes io."""
        fio = PickleIO()
        bio = BytesIO()
        pickle.dump(random_patch, bio)
        bio.seek(0)
        fmt = fio.get_format(bio)
        assert "PICKLE" in fmt
        spool = fio.read(bio)
        assert spool[0] == random_patch


class TestScan:
    """Tests for scanning pickle files/."""

    comp_attrs = (
        "data_type",
        "data_units",
        "time_step",
        "time_min",
        "time_max",
        "distance_min",
        "distance_max",
        "distance_step",
        "tag",
        "network",
    )

    def test_scan_attrs_eq_read_attrs(self, pickle_patch_path):
        """Ensure read/scan produce the same attrs."""
        scan_list = dc.scan(pickle_patch_path)
        patch_attrs_list = [x.attrs for x in dc.read(pickle_patch_path)]

        for scan_attrs, patch_attrs in zip(scan_list, patch_attrs_list):
            for attr in self.comp_attrs:
                scan_attr = getattr(scan_attrs, attr)
                patch_attr = getattr(patch_attrs, attr)
                assert scan_attr == patch_attr



"""Dispersion computation using the phase-shift (Park et al., 1999) method."""

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import numpy.fft as nft

from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.utils.patch import patch_function


@patch_function(required_dims=("time", "distance"))
def dispersion_phase_shift(
    patch: PatchType,
    phase_velocities: Sequence[float],
    approx_resolution: None | float = None,
    approx_freq: None | tuple[float, float] = None,
) -> PatchType:
    """
    Compute dispersion images using the phase-shift method.

    Parameters
    ----------
    patch
        Patch to transform. Has to have dimensions of time and distance.
        It also needs to be right-sided (see notes below).
    phase_velocities
        NumPY array of positive velocities, monotonically increasing, for
        which the dispersion will be computed.
    approx_resolution
        Approximated frequency (Hz) resolution for the output. If left empty,
        the frequency resolution is dictated by the number of samples.
    approx_freq
        Minimum and maximum frequency to compute dispersion for, in Hz
        If left empty, minimum is 0 Hz, and maximum is Nyquist

    Notes
    -----
    - See also @park1998imaging

    - Inspired by https://geophydog.cool/post/masw_phase_shift/.

    - Dims/Units of the output are forced to be 'frequency' ('Hz')
      and 'velocity' ('m/s').

    - The patch's distance coordinates are assumed to be ordered by
    distance from the source, and not "fiber distance". In other
    words, data are effectively mapped along a 2-D line.

    - The input shot gather must be right-sided meaning the
    wavefield propagates from lower to higher channel numbers.
    Always plot the patch first to verify its orientation.
    If the gather is left-sided, simply mirror the patch along
    the distance axis (see Example 2 below).

    Examples
    --------
    ```{python}
    import dascore as dc
    import numpy as np

    # Example 1 - Right-sided wavefield
    patch = (
        dc.get_example_patch('dispersion_event')
    )

    disp_patch = patch.dispersion_phase_shift(np.arange(100,1500,1),
                approx_resolution=0.1,approx_freq=[5,70])
    ax = disp_patch.viz.waterfall(show=False,cmap=None)
    ax.set_xlim(5, 70)
    ax.set_ylim(1500, 100)
    disp_patch.viz.waterfall(show=True, ax=ax)

    # Example 2 - Left-sided wavefield
    patch = (
        dc.get_example_patch('dispersion_event')
    )
    axis = patch.dims.index("distance")
    flipped_data = np.flip(patch.data, axis=axis)
    mirrored_patch = patch.update(data=flipped_data)

    disp_patch = mirrored_patch.dispersion_phase_shift(np.arange(100,1500,1),
            approx_resolution=0.1,approx_freq=[5,70])
    ```
    """
    patch_cop = patch.convert_units(distance="m").transpose("distance", "time")
    dist = patch_cop.coords.get_array("distance")
    time = patch_cop.coords.get_array("time")

    dt = (time[1] - time[0]) / np.timedelta64(1, "s")

    if not np.all(np.diff(phase_velocities) > 0):
        raise ParameterError(
            "Velocities for dispersion must be monotonically increasing"
        )

    if np.amin(phase_velocities) <= 0:
        raise ParameterError("Velocities must be positive.")

    if approx_resolution is not None and approx_resolution <= 0:
        raise ParameterError("Frequency resolution has to be positive")

    if not approx_freq:
        approx_min_freq = 0
        approx_max_freq = 0.5 / dt
    else:
        approx_min_freq = approx_freq[0]
        approx_max_freq = approx_freq[1]
        if approx_min_freq <= 0 or approx_max_freq <= 0:
            msg = "Minimal and maximal frequencies have to be positive"
            raise ParameterError(msg)

        if approx_min_freq >= approx_max_freq:
            msg = "Maximal frequency needs to be larger than minimal frequency"
            raise ParameterError(msg)

        if approx_min_freq >= 0.5 / dt or approx_max_freq >= 0.5 / dt:
            msg = "Frequency range cannot exceed Nyquist"
            raise ParameterError(msg)

    nchan = dist.size
    nt = time.size
    assert (nchan, nt) == patch_cop.data.shape

    fs = 1 / dt
    if approx_resolution is not None:
        approxnf = int(nt * (fs / (nt)) / approx_resolution)
        f = np.arange(approxnf) * fs / (approxnf - 1)
    else:
        f = np.arange(nt) * fs / (nt - 1)

    nf = np.size(f)

    nv = np.size(phase_velocities)
    w = 2 * np.pi * f
    fft_d = np.zeros((nchan, nf), dtype=complex)
    for i in range(nchan):
        fft_d[i] = nft.fft(patch_cop.data[i, :], n=nf)

    fft_d = np.divide(
        fft_d, abs(fft_d), out=np.zeros_like(fft_d), where=abs(fft_d) != 0
    )

    fft_d[np.isnan(fft_d)] = 0

    first_live_f = np.argmax(w >= 2 * np.pi * approx_min_freq)
    last_live_f = np.argmax(w >= 2 * np.pi * approx_max_freq)
    w = w[first_live_f:last_live_f]
    fft_d = fft_d[:, first_live_f:last_live_f]
    nlivef = last_live_f - first_live_f

    if nlivef < 1:
        msg = "Combination of frequency resolution and range is not an array"
        raise ParameterError(msg)

    fc = np.zeros(shape=(nv, nlivef))
    preamb = 1j * np.outer(dist, w)
    for ci in range(nv):
        fc[ci, :] = abs(sum(np.exp(preamb / phase_velocities[ci]) * fft_d))

    attrs = patch.attrs.update(category="dispersion")
    coords = dict(velocity=phase_velocities, frequency=w / (2 * np.pi))

    disp_patch = patch.new(
        data=fc / nchan, coords=coords, attrs=attrs, dims=["velocity", "frequency"]
    )
    return disp_patch.set_units(velocity="m/s", frequency="Hz")



"""IO module for reading and writing RSF file format support of MADAGASCAR."""
# from __future__ import annotations

# import segyio
import datetime as dt
from pathlib import Path

import numpy as np

import dascore as dc
from dascore.io.core import FiberIO
from dascore.utils.time import to_float

RSFKEYS_WRITE = ("in", "esize", "data_format")


class RSFV1(FiberIO):
    """An IO class supporting the RSF format of MADAGASCAR."""

    name = "rsf"
    preferred_extensions = "rsf"
    # also specify a version so when version 2 is released you can
    # just make another class in the same module named rsfV2.
    version = "1"

    def write(self, spool, path, data_path=None, **kwargs):
        """
        Write a patch to RSF format.

        if no data_path is NOT specified, the header and binary will be
        packed together
        data_path needs to be bindata_file.rsf or /location/of/bindata_file.rsf
        (NO '@')

        path needs to be hdr_file.rsf or /location/of/hdr_file.rsf

        spool needs to have a single patch in it

        time origin is forced to 0.0 in the new RSF file, but a dummy
        variable named 'starttime' still holds the real start time

        Parameters
        ----------
        spool
            The input spool to convert to rsf, must have exactly one patch.
        path
            Path to create the rsf file
        data_path
            If data and rsf header information are to be separate, the
            location of the data file. Needs to be bindata_file.rsf or
            /location/of/bindata_file.rsf (no '@')

        Notes
        -----
        - Patch datatype is converted to float32 for compatibility with
        Madagascar (may be able to keep dytpe in the future)
        """
        assert len(spool) == 1
        patch = spool[0]
        axis_lengs = patch.shape
        axis_origs = [to_float(patch.get_coord(x).start) for x in patch.dims]
        axis_steps = [to_float(patch.get_coord(x).step) for x in patch.dims]
        axis_names = patch.dims
        axis_units = [patch.get_coord(x).units.units for x in patch.dims]

        data = patch.data
        dtype = np.dtype(data.dtype)

        if not (np.issubdtype(dtype, np.integer) or np.issubdtype(dtype, np.floating)):
            raise ValueError("Data format is not integer or floating.")
        ## we are casting to float32 to ensure that esize=4 for m8r
        data = patch.data.astype(np.float32)
        dtype = np.dtype(data.dtype)
        file_esize = dtype.itemsize
        file_formt = 'data_format="native_float"'

        data_bytes = data.astype(np.float32).tobytes("F")

        hdr_str = f"DASCORE {dc.__version__}   {dt.datetime.now()} \n"

        length = len(axis_lengs)
        hdr_info = [hdr_str, file_formt, f"esize={file_esize}"]
        for i in range(length):
            hdr_info.append(f"n{i+1}={axis_lengs[i]}")
            if axis_names[i] == "time":
                hdr_info.append(f"o{i+1}=0.0")
                hdr_info.append(f"starttime={axis_origs[i]}")
            else:
                hdr_info.append(f"o{i+1}={axis_origs[i]}")
            hdr_info.append(f"d{i+1}={axis_steps[i]}")
            hdr_info.append(f'label{i+1}="{axis_names[i]}"')
            hdr_info.append(f'unit{i+1}="{axis_units[i]}"')

        if data_path is not None:
            # outputs header and binary separately (.rsf and .rsf@)
            outdatapath = Path(str(data_path) + "@")
            outdatapath.parent.mkdir(exist_ok=True, parents=True)
            hdr_info.append(f'in="{data_path}@"')
            with outdatapath.open("wb") as fi:
                fi.write(data_bytes)
            out = "\n".join(hdr_info)
            outpath = Path(str(path))
            outpath.parent.mkdir(exist_ok=True, parents=True)
            with outpath.open("w") as fi:
                fi.write(out)
        else:
            # outputs header and binary combined (.rsf with both hdr and bin)
            hdr_info.append('in="stdin"\n\n')
            # hdr_info.append(data)
            out = "\n".join(hdr_info)
            outpath = Path(str(path))
            outpath.parent.mkdir(exist_ok=True, parents=True)
            with outpath.open("w") as fi:
                fi.write(out)
            with outpath.open("ab") as fi:
                fi.write(data_bytes)



"""Utilities for terra15."""

from __future__ import annotations

import dascore as dc
from dascore.constants import timeable_types
from dascore.core import Patch
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_coord
from dascore.utils.misc import maybe_get_items
from dascore.utils.time import to_datetime64, to_timedelta64

# --- Getting format/version


def _get_terra15_version_str(hdf_fi) -> str:
    """Return the version string for terra15 file."""
    # define a few root attrs that act as a "fingerprint" for terra15 files
    expected_attrs = {
        "acoustic_bandwidth_end",
        "amplifier_incoming_current",
        "file_start_computer_time",
        "file_version",
    }
    root_attrs = hdf_fi.attrs
    if not set(root_attrs).issuperset(expected_attrs):
        return ""
    return str(root_attrs["file_version"])


# --- Getting File summaries


def _get_time_node(data_node):
    """
    Get the time node from data.

    This will prefer GPS time but gets posix time if it is missing.
    """
    try:
        time = data_node["gps_time"]
    except (IndexError, KeyError):
        time = data_node["posix_time"]
    return time


def _get_scanned_time_info(data_node):
    """Get the min, max, len, and dt from time array."""
    time = _get_time_node(data_node)
    t_len = len(time)
    # first try fast path by tacking first/last of time
    tmin, tmax = time[0], time[-1]
    # This doesn't work if an incomplete datablock exists at the end of
    # the file. In this case we need to read/filter time array (slower).
    if tmin > tmax:
        time = time[:]
        time_filtered = time[time > 0]
        t_len = len(time_filtered)
        tmin, tmax = time_filtered[0], time_filtered[-1]
    # surprisingly, using gps time column, dt is much different than dt
    # reported in data attrs so we calculate it this way.
    dt = (tmax - tmin) / (t_len - 1)
    starttime = to_datetime64(tmin)
    time_step = to_timedelta64(dt)
    endtime = starttime + time_step * (t_len - 1)
    return starttime, endtime, t_len, time_step


def _get_version_data_node(root):
    """Get the version, time, and data node from terra15 file."""
    version = str(root.attrs["file_version"])
    if version == "4":
        data_type = root.attrs["data_product"]
        data_node = root[data_type]
    elif version in {"5", "6"}:
        data_node = root["data_product"]
    else:
        raise NotImplementedError("Unknown Terra15 version")
    return version, data_node


def _scan_terra15(h5_fi, data_node, extras=None):
    """Scan a terra15 file, return metadata."""
    out = extras
    out.update(_get_default_attrs(h5_fi.attrs))
    coords = {
        "time": _get_time_coord(data_node, snap_dims=True),
        "distance": _get_distance_coord(h5_fi),
    }
    out["coords"] = coords
    return [dc.PatchAttrs(**out)]


# --- Reading patch


def _get_raw_time_coord(data_node):
    """Read the time from the data node and return it."""
    time = _get_time_node(data_node)[:]
    return get_coord(data=to_datetime64(time))


def _read_terra15(
    pyfi,
    time: tuple[timeable_types, timeable_types] | None = None,
    distance: tuple[float, float] | None = None,
    snap_dims: bool = True,
) -> Patch:
    """
    Read a terra15 file.

    Notes
    -----
    The time array is complicated. There is GPS time and Posix time included
    in the file. In version 0.0.6 and less of dascore we just used gps time.
    However, sometimes this results in subsequent samples having a time before
    the previous sample (time did not increase monotonically).

    So now, we use the first GPS sample, the last sample, and length
    to determine the dt (new in dascore>0.0.11).
    """
    _, data_node = _get_version_data_node(pyfi)
    time_coord_ = _get_time_coord(data_node, snap_dims)
    time_coord, time_slice = time_coord_.select(time)
    time_len = len(time_coord)
    # get data and sliced distance coord
    dist_coord = _get_distance_coord(pyfi)
    dist_coord, dist_slice = dist_coord.select(distance)
    _data = data_node["data"]
    # checks for incomplete data blocks
    if _data.shape[0] > time_len:
        new_start = time_slice.start or 0
        t_stop = time_slice.stop
        new_stop = new_start + time_len if t_stop is None else t_stop
        time_slice = slice(new_start, new_stop)
    data = data_node["data"][time_slice, dist_slice]
    coords = get_coord_manager(
        {"time": time_coord, "distance": dist_coord},
        dims=("time", "distance"),
    )
    dims = ("time", "distance")
    attrs = _get_default_attrs(pyfi.attrs)
    attrs["coords"] = coords
    return Patch(data=data, coords=coords, attrs=attrs, dims=dims)


def _get_default_attrs(root_node_attrs):
    """
    Return the required/default attributes which can be fetched from attributes.

    Note: missing time, distance absolute ranges. Downstream functions should handle
    this.
    """
    out = {}
    _root_attrs = {
        "data_product": "data_type",
        "serial_number": "instrument_id",
        "data_product_units": "data_units",
        "pulse_rate": "pulse_rate",
        "pulse_length": "pulse_length",
        "gauge_length": "gauge_length",
    }
    out.update(maybe_get_items(root_node_attrs, _root_attrs))
    return out


def _get_time_coord(data_node, snap_dims=True):
    """Get the time coordinate."""
    t_min, t_max, time_len, d_time = _get_scanned_time_info(data_node)
    if snap_dims:
        kwargs = dict(start=t_min, stop=t_max + d_time, step=d_time, units="s")
        time_coord = get_coord(**kwargs)
    else:
        time_coord = _get_raw_time_coord(data_node)
    return time_coord


def _get_distance_coord(root):
    """Get the distance coordinate."""
    # TODO: At least for the v4 test file, sensing_range_start, sensing_range_stop,
    # nx, and dx are not consistent, meaning d_min + dx * nx != d_max
    # so I just used this method. We need to look more into this.
    attrs = getattr(root, "attrs", root)
    start, step = attrs["sensing_range_start"], attrs["dx"]
    stop = attrs["sensing_range_start"] + (attrs["nx"] - 1) * attrs["dx"]
    return get_coord(start=start, stop=stop + step, step=step, units="m")



"""Module for re-sampling patches."""

from __future__ import annotations

from typing import Literal

import numpy as np
from scipy.signal import decimate as scipy_decimate

import dascore as dc
import dascore.compat as compat
from dascore.constants import PatchType
from dascore.exceptions import FilterValueError
from dascore.units import get_filter_units
from dascore.utils.patch import (
    get_dim_axis_value,
    get_start_stop_step,
    patch_function,
)
from dascore.utils.time import to_int, to_timedelta64


def _apply_scipy_decimation(patch, factor, ftype, axis):
    """
    Apply decimation along axis.
    """
    try:
        data = scipy_decimate(patch.data, factor, ftype=ftype, axis=axis)
    except ValueError as e:
        msg = (
            "Scipy decimation failed. This can happen for dimensions with"
            "few elements. Consider setting filter_type to False. The raised "
            f"exception was {e}"
        )
        raise FilterValueError(msg)
    return data


@patch_function()
def decimate(
    patch: PatchType,
    filter_type: Literal["iir", "fir", None] = "iir",
    copy=True,
    **kwargs,
) -> PatchType:
    """
    Decimate a patch along a dimension.

    Parameters
    ----------
    filter_type
        filter type to use to avoid aliasing. Options are:
            iir - infinite impulse response
            fir - finite impulse response
            None - No pre-filtering, not recommended, may cause aliasing
    copy
        If True, copy the decimated data array. This is needed if you want
        the old array to get gc'ed to free memory otherwise a view is returned.
        Only applies when filter_type == None.
    **kwargs
        Used to pass dimension and factor. For example `time=10` is 10x
        decimation along the time axis.

    Notes
    -----
    - Simply uses scipy.signal.decimate if filter_type is specified.
      Otherwise,just slice data long specified dimension only including
      every n samples.

    - If the decimation dimension is small, this can fail due to lack of
      padding values.

    Examples
    --------
    # Simple example using iir
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> decimated_irr = patch.decimate(time=10, filter_type='iir')
    >>> # Example using fir along distance dimension
    >>> decimated_fir = patch.decimate(distance=10, filter_type='fir')
    """
    dim, axis, factor = get_dim_axis_value(patch, kwargs=kwargs)[0]
    coords, slices = patch.coords.decimate(**{dim: int(factor)})
    # Apply scipy.signal.decimate and get new coords
    if filter_type:
        data = _apply_scipy_decimation(patch.data, factor, ftype=filter_type, axis=axis)
    else:  # No filter, simply slice along specified dimension.
        data = patch.data[slices]
        # Need to copy so array isn't a slice and holds onto reference of parent
        data = np.array(data) if copy else data
    # Update delta_dim since spacing along dimension has changed.
    return patch.new(data=data, coords=coords)


@patch_function()
def interpolate(patch: PatchType, kind: str | int = "linear", **kwargs) -> PatchType:
    """
    Set coordinates of patch along a dimension using interpolation.

    Parameters
    ----------
    patch
        The patch object to which interpolation is applied.
    kind
        The type of interpolation. See Notes for more details.
        If a string, the following are supported:
            linear - linear interpolation between a pair of points.
            nearest - use the nearest sample for interpolation.
        If an int, it specifies the order of spline to use. EG 1 is a linear
            spline, 2 is quadratic, 3 is cubic, etc.

    **kwargs
        Used to specify dimension and interpolation values. Use a value of
        None to "snap" coordinate to evenly sampled points along coordinate.

    Notes
    -----
    This function just uses scipy's interp1d function under the hood.
    See scipy.interpolate.interp1d for information.

    See also [snap](`dascore.core.Patch.snap_coords`).

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # up-sample time coordinate
    >>> time = patch.coords.get_array('time')
    >>> new_time = np.arange(time.min(), time.max(), 0.5*patch.attrs.time_step)
    >>> patch_uptime = patch.interpolate(time=new_time)
    >>> # interpolate unevenly sampled dim to evenly sampled
    >>> patch = dc.get_example_patch("wacky_dim_coords_patch")
    >>> patch_time_even = patch.interpolate(time=None)
    """
    dim, axis, samples = get_dim_axis_value(patch, kwargs=kwargs)[0]
    # if samples is None, get evenly sampled coords along dimension.
    if samples is None:
        coord = patch.coords.coord_map[dim]
        samples = coord.snap().values
    # we need to make sure only real numbers are used, interp1d doesn't support
    # datetime64 yet.
    coord_num = to_int(patch.coords.get_array(dim))
    samples_num = to_int(samples)
    func = compat.interp1d(
        coord_num, patch.data, axis=axis, kind=kind, fill_value="extrapolate"
    )
    out = func(samples_num)
    cm = patch.coords
    associated_dims = cm.dim_map[dim]
    coord_new = dc.core.get_coord(data=samples)
    cm_new = cm.update(**{dim: (associated_dims, coord_new)})
    return patch.new(data=out, coords=cm_new)


@patch_function()
def resample(
    patch: PatchType, window=None, interp_kind="linear", samples=False, **kwargs
) -> PatchType:
    """
    Resample along a single dimension using Fourier Method and interpolation.

    The dimension which should be resampled is passed as kwargs. The key
    is the dimension name and the value is the new sampling period.

    Since Fourier methods only support adding or removing an integer number
    of frequency bins, the exact desired sampling rate is often not achievable
    with resampling alone. If the fourier resampling doesn't produce the exact
    result, an interpolation (see [interpolate](`dascore.proc.interpolate`))
    is used to achieve the desired sampling rate.

    Parameters
    ----------
    patch
        The patch to resample.
    window
        The Fourier-domain window that tapers the Fourier spectrum. See
        scipy.signal.resample for details. Only used if method == 'fft'.
    interp_kind
        The interpolation type if output of fourier resampling doesn't produce
        exactly the right sampling rate.
    samples
        If true, the values in kwargs represent the number of samples along
        The specified dimension.
    **kwargs
        keyword arguments to specify dimension and new sampling value. Units
        can also be used to specify sampling_period or frequency.

    Notes
    -----
    - Unless `samples` is `True`, this function requires a sampling_period.
    - The resulting Patch can be slightly shorter than the input Patch.

    Examples
    --------
    >>> # resample a patch along time dimension to 10 ms
    >>> import numpy as np
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> new = patch.resample(time=np.timedelta64(10, 'ms'))
    >>> # Resample time dimension to 50 Hz
    >>> from dascore.units import Hz
    >>> new = patch.resample(time=(50 * Hz))
    >>> # Resample distance dimension to a sampling period of 15m
    >>> from dascore.units import m
    >>> new = patch.resample(distance=15 * m)
    >>> # Resample time axis such that there are 50 samples total
    >>> new = patch.resample(time=50, samples=True)

    See Also
    --------
    [decimate](`dascore.proc.resample.decimate`)
    [interpolate](`dascore.proc.resample.interpolate`)
    """
    dim, axis, value = get_dim_axis_value(patch, kwargs=kwargs)[0]
    coord = patch.get_coord(dim, require_sorted=True, require_evenly_sampled=True)
    new_step = None
    if not samples:
        step = coord.step
        coord_units = dc.get_quantity(coord.units)
        # inverse coord unit to trick filter units into giving correct units.
        if coord_units is not None:
            coord_units = 1 / coord_units
        new_step, _ = get_filter_units(value, value, to_unit=coord_units)
        # nasty hack so that ints/floats get converted to seconds.
        if isinstance(step, np.timedelta64):
            new_step = to_timedelta64(new_step)
        current_sig_len = patch.data.shape[axis]
        new_len = current_sig_len * (step / new_step)
    else:
        new_len = value
    # do the resampling
    data, new_coord = compat.resample(
        patch.data, int(np.round(new_len)), t=coord, axis=axis, window=window
    )
    cm = patch.coords.update(**{dim: new_coord})
    out = patch.new(data=data, coords=cm)
    # Interpolate if new sampling rate is not very close to desired sampling rate.
    if not samples and not np.isclose(new_len, np.round(new_len)):
        start, stop, step = get_start_stop_step(out, dim)
        new_coord = np.arange(start, stop, new_step)
        out = interpolate(out, kind=interp_kind, **{dim: new_coord})
    return out



"""Utilities for working with the Patch class."""

from __future__ import annotations

import functools
import inspect
import sys
import warnings
from collections import namedtuple
from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal

import numpy as np
import pandas as pd
import pydantic
from pydantic import TypeAdapter

import dascore as dc
from dascore.constants import (
    DEFAULT_ATTRS_TO_IGNORE,
    FLOAT_PRECISION,
    WARN_LEVELS,
    PatchType,
    SpoolType,
    check_behavior_description,
    dascore_styles,
)
from dascore.exceptions import (
    CoordDataError,
    IncompatiblePatchError,
    ParameterError,
    PatchAttributeError,
    PatchCoordinateError,
)
from dascore.units import get_quantity
from dascore.utils.attrs import combine_patch_attrs
from dascore.utils.coordmanager import merge_coord_managers
from dascore.utils.docs import compose_docstring
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import (
    _apply_union_indexers,
    _merge_tuples,
    all_diffs_close_enough,
    get_middle_value,
    iterate,
    to_object_array,
    warn_or_raise,
    yield_sub_sequences,
)
from dascore.utils.time import to_float

attr_type = dict[str, Any] | str | Sequence[str] | None

_DimAxisValue = namedtuple("DimAxisValue", ["dim", "axis", "value"])


def _format_values(val):
    """String formatting for values for history string."""
    if isinstance(val, list | tuple):
        out = ",".join(_format_values(x) for x in val)
        out = f"({out})" if isinstance(val, tuple) else f"[{out}]"
    elif isinstance(val, np.ndarray):
        # make sure numpy strings arent too long!
        out = np.array2string(
            val,
            precision=FLOAT_PRECISION,
            threshold=dascore_styles["patch_history_array_threshold"],
        )
    else:
        out = str(val)
    return out


def _func_and_kwargs_str(func: Callable, patch, *args, **kwargs) -> str:
    """Get a str rep of the function and input args."""
    callargs = inspect.getcallargs(func, patch, *args, **kwargs)
    callargs.pop("patch", None)
    callargs.pop("self", None)
    kwargs_ = callargs.pop("kwargs", {})
    arguments = []
    arguments += [
        f"{k}={_format_values(v)!r}" for k, v in callargs.items() if v is not None
    ]
    arguments += [
        f"{k}={_format_values(v)!r}" for k, v in kwargs_.items() if v is not None
    ]
    arguments.sort()
    out = f"{func.__name__}("
    if arguments:
        out += f"{','.join(arguments)}"
    return out + ")"


def _get_history_str(patch: PatchType, func, *args, _history="full", **kwargs) -> str:
    """
    Log history of a function being called on a Patch object.

    Parameters
    ----------
    patch
        The patch which will track history.
    func
        A callable which takes the patch as the first argument.
    *args
        The arguments passed to the function.
    _history
        String specifying how the history is to be recorded.
    **kwargs
        kwargs for func.
    """
    if _history is None:
        return None
    if _history == "full":
        history_str = _func_and_kwargs_str(func, patch, *args, **kwargs)
    else:
        history_str = str(func.__name__)
    return history_str


def check_patch_coords(
    patch: PatchType,
    dims: Sequence[str] | None = None,
    coords: Sequence[str] | None = None,
):
    """
    Check if a patch object has required coordinates, else raise PatchDimError.

    Parameters
    ----------
    patch
        The input patch
    dims
        A tuple of required dimension names.
    coords
        A tuple of required coordinate names.
    """
    missing = set()
    if dims is not None:
        dim_set = set(patch.dims)
        missing |= set(dims) - dim_set
    if coords is not None:
        coord_set = set(patch.coords.coord_map)
        missing |= set(coords) - coord_set
    if missing:
        msg = f"patch is missing required coordinates: {tuple(missing)}"
        raise PatchCoordinateError(msg)
    return patch


def check_patch_attrs(patch: PatchType, required_attrs: attr_type) -> PatchType:
    """
    Check for expected attributes.

    Parameters
    ----------
    patch
        The patch to validate
    required_attrs
        The expected attrs. Can be a sequence or mapping. If sequence, only
        check that attrs exist. If mapping also check that values are equal.
    """
    if required_attrs is None:
        return patch
    # test that patch attr mapping is equal
    patch_attrs = patch.attrs
    if isinstance(required_attrs, Mapping):
        sub = {i: patch_attrs[i] for i in required_attrs}
        if sub != dict(required_attrs):
            msg = f"Patch's attrs {sub} are not required attrs: {required_attrs}"
            raise PatchAttributeError(msg)
    else:
        missing = set(required_attrs) - set(dict(patch.attrs))
        if missing:
            msg = f"Patch is missing the following attributes: {missing}"
            raise PatchAttributeError(msg)
    return patch


def patch_function(
    required_dims: tuple[str, ...] | Callable | None = None,
    required_coords: tuple[str, ...] | None = None,
    required_attrs: attr_type = None,
    history: Literal["full", "method_name", None] = "full",
    validate_call: bool = False,
):
    """
    Decorator to mark a function as a patch method.

    Parameters
    ----------
    required_dims
        A tuple of dimensions which must be found in the Patch.
    required_coords
        A tuple of coordinates which must be found in the Patch.
    required_attrs
        A dict of attributes which must be found in the Patch and whose
        values must be equal to those provided.
    history
        Specifies how to track history on Patch.
            Full - Records function name and str version of input arguments.
            method_name - Only records method name. Useful if args are long.
            None - Function call is not recorded in history attribute.
    validate_call
        If True, use pydantic to validate the function call. This can save
        quite a lot of code in validation checks, but does have some overhead.
        See [validate_call](https://docs.pydantic.dev/latest/api/validate_call/).

    Examples
    --------
    >>> import dascore as dc
    >>>
    >>> # 1. A patch method which requires dimensions (time, distance)
    >>> @dc.patch_function(required_dims=('time', 'distance'))
    ... def do_something(patch):
    ...     ...   # raises a PatchCoordsError if patch doesn't have time,
    ...     #  distance
    >>>
    >>> # 2. A patch method which requires an attribute 'data_type' == 'DAS'
    >>> @dc.patch_function(required_attrs={'data_type': 'DAS'})
    ... def do_another_thing(patch):
    ...     ...  # raise PatchAttributeError if patch doesn't have attribute
    ...     # called "data_type" or its values is not equal to "DAS".
    >>>
    >>> # 3. A patch method which does type checking on inputs.
    >>> # The `Field` instance can require various data properties (like ranges)
    >>> from typing_extensions import Annotated, Literal
    >>> from pydantic import Field
    >>> @dc.patch_function(validate_call=True)
    ... def do_type_thing(
    ...     patch,
    ...     int_le_10_ge_1: int = Field(ge=1, le=10, default=1),
    ...     option: Literal["min", "max", None] = None,
    ... ):
    ...     ...

    Notes
    -----
    - The original function can still be accessed with the raw_function
      attribute. This may be useful for avoiding calling the patch_func
      machinery multiple times from within another patch function.

    - If using `PatchType` or `SpoolType` type variables from the
      [constants module](`dascore.constants`), make sure dascore is imported
      as dc at the top of the file where the patch function is defined so
      the forward refs can be resolved properly for type checking.
    """

    def _wrapper(func):
        if validate_call:
            config = dict(arbitrary_types_allowed=True)
            func = pydantic.validate_call(func, config=config)

        @functools.wraps(func)
        def _func(patch, *args, **kwargs):
            check_patch_coords(
                patch,
                dims=required_dims,
                coords=required_coords,
            )
            check_patch_attrs(patch, required_attrs)
            out: PatchType = func(patch, *args, **kwargs)
            # attach history string. Need to consider something a bit less hacky.
            if out is not patch and hasattr(out, "attrs"):
                hist_str = _get_history_str(
                    patch, func, *args, _history=history, **kwargs
                )
                if hist_str:
                    hist = list(out.attrs.history)
                    hist.append(hist_str)
                    out = out.update_attrs(history=hist)
            return out

        # Attach original function. Although we want to encourage raw_function
        # for consistency with pydantic, we leave this to not break old code.
        _func.func = getattr(func, "raw_function", func)
        # matches pydantic naming.
        _func.raw_function = getattr(func, "raw_function", func)
        _func.__wrapped__ = func

        return _func

    if callable(required_dims):  # the decorator is used without parens
        return patch_function()(required_dims)

    return _wrapper


def patches_to_df(
    patches: Sequence[PatchType] | SpoolType | pd.DataFrame,
) -> pd.DataFrame:
    """
    Return a dataframe.

    Parameters
    ----------
    patches
        A sequence of :class:`dascore.Patch`

    Returns
    -------
    A dataframe with the attrs of each patch converted to a columns
    plus a field called 'patch' which contains a reference to the patches.
    """
    # Handle spool case
    if hasattr(patches, "get_contents"):
        df = patches.get_contents()
    elif isinstance(patches, pd.DataFrame):
        df = patches
    else:
        df = dc.scan_to_df(
            patches,
            exclude=(),
        )
        if df.empty:  # create empty df with appropriate columns
            cols = list(dc.PatchAttrs().model_dump())
            df = pd.DataFrame(columns=cols).assign(patch=None, history=None)
        else:  # else populate with patches and concat history
            history = df["history"].apply(lambda x: ",".join(x))
            df = df.assign(patch=to_object_array(patches), history=history)
    # Ensure history is in df
    if "history" not in df.columns:
        df = df.assign(history="")
    if "patch" not in df.columns:
        df["patch"] = to_object_array(patches)
    return df


def merge_patches(
    patches: Sequence[PatchType] | pd.DataFrame | SpoolType,
    dim: str = "time",
    check_history: bool = True,
    tolerance: float = 1.5,
) -> Sequence[PatchType]:
    """
    Merge all compatible patches in spool or patch list together.

    Parameters
    ----------
    patches
        A sequence of patches to merge (if compatible)
    dim
        The dimension along which to merge
    check_history
        If True, only merge patches with common history. This will, for
        example, prevent merging filtered and unfiltered data together.
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. e.g., the default value means any patches
        with gaps <= 1.5 * dt will be merged.
    """
    msg = (
        "merge_patches is deprecated. Use spool.chunk instead. "
        "For example, to merge a list of patches you can use: "
        "dascore.spool(patch_list).chunk(time=None) to merge on the time "
        "dimension."
    )
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    return dc.spool(patches).chunk(**{dim: None}, tolerance=tolerance)


def _force_patch_merge(patch_dict_list, merge_kwargs, **kwargs):
    """
    Force a merge of the patches along a dimension.

    This function is used in conjunction with `spool.chunk`, which
    does all the compatibility checks beforehand.
    """

    def _get_merge_col(df):
        dims = df["dims"].unique()
        assert len(dims) == 1
        dims = dims[0].split(",")
        dims_vary = pd.Series({x: False for x in dims})
        for dim in dims:
            cols = [f"{dim}_min", f"{dim}_max", f"{dim}_step"]
            vals = df[cols].values
            vals_eq = vals == vals[[0], :]
            vals_null = pd.isnull(vals)
            columns_equal = (vals_eq | vals_null).all(axis=1)
            dims_vary[dim] = not np.all(columns_equal)
        assert dims_vary.sum() <= 1, "Only one dimension can vary for forced merge"
        if not dims_vary.any():  # the case of complete overlap.
            return None
        return dims_vary[dims_vary].index[0]

    def _maybe_step(df, dim):
        """Get the expected step if all steps are close, else None."""
        col = df[f"{dim}_step"].values
        if all_diffs_close_enough(col):
            return get_middle_value(col)
        return None

    def _get_new_coord(df, merge_dim, coords):
        """Get new coordinates, also validate anticipated sampling."""
        new_coord = merge_coord_managers(coords, dim=merge_dim)
        expected_step = _maybe_step(df, merge_dim)
        if not pd.isnull(expected_step):
            new_coord = new_coord.snap(merge_dim)[0]
            # TODO slightly different dt can be produced, let pass for now
            # need to think more about how the merging should work.
        return new_coord

    df = pd.DataFrame(patch_dict_list)
    merge_dim = _get_merge_col(df)
    merge_kwargs = merge_kwargs if merge_kwargs is not None else {}
    if merge_dim is None:  # nothing to merge, complete overlap
        return [patch_dict_list[0]]
    dims = df["dims"].iloc[0].split(",")
    # get patches, ensure they are oriented the same.
    patches = [x.transpose(*dims) for x in df["patch"]]
    axis = patches[0].dims.index(merge_dim)
    # get data, coords, attrs for merging patch together.
    datas = [x.data for x in patches]
    coords = [x.coords for x in patches]
    attrs = [x.attrs for x in patches]
    new_data = np.concatenate(datas, axis=axis)
    new_coord = _get_new_coord(df, merge_dim, coords)
    coord = new_coord.coord_map[merge_dim] if merge_dim in dims else None
    new_attrs = combine_patch_attrs(attrs, merge_dim, coord=coord, **merge_kwargs)
    patch = dc.Patch(data=new_data, coords=new_coord, attrs=new_attrs, dims=dims)
    new_dict = {"patch": patch}
    return [new_dict]


def get_start_stop_step(patch: PatchType, dim):
    """Convenience method for getting start, stop, step for a given coord."""
    assert dim in patch.dims, f"{dim} is not in Patch dimensions of {patch.dims}"
    coord = patch.get_coord(dim)
    start = coord.min()
    stop = coord.max()
    step = coord.step
    return start, stop, step


def get_patch_names(
    patch_data: pd.DataFrame | dc.Patch | dc.BaseSpool,
    prefix="DAS",
    attrs=("network", "station", "tag"),
    coords=("time",),
    sep="__",
    strip_extension=True,
) -> pd.Series:
    """
    Generates the default name of patch data.

    Parameters
    ----------
    patch_data
        A container with patch data.
    prefix
        A string to prefix the names.
    attrs
        The Patch attrs to include in the name.
    coords
        The coordinate ranges to use in the names.
    sep
        The separator for each value.
    strip_extension
        If True, remove extensions when getting name from a file path.
        See the notes section for more details.

    Notes
    -----
    There are two special cases where the default logic is overwritten.
    The first one, is when a column called "name" already exists. This
    will simply be returned.

    The second is when a column called "path" exists. In this case, the
    output will be the file name with the extension removed (if
    strip_extension). The path must use '/' as a delinater.

    See Also
    --------
    - [`Patch.get_patch_name`](`dascore.Patch.get_patch_name`)
    - [`Spool.get_patch_names`](`dascore.BaseSpool.get_patch_names`)

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.patch import get_patch_names
    >>>
    >>> # Get a series of names from a patch or spool
    >>> patch = dc.get_example_patch()
    >>> spool = dc.get_example_spool()
    >>> patch_name = get_patch_names(patch)
    >>> spool_name = get_patch_names(patch)
    >>>
    >>> # Use the Patch/Spool methods
    >>> spool_series = spool.get_patch_names()
    >>> patch_name = patch.get_patch_name() # a str w/ name.
    """

    def _format_time_column(ser):
        """Format the time column."""
        ser = ser.astype(str).str.split(".", expand=True)[0]
        chars_to_replace = (":", "-")
        for char in chars_to_replace:
            ser = ser.str.replace(char, "_")
        ser = ser.str.replace(" ", "T")
        return ser

    def _format_time_columns(df):
        """Format the datetime string in a sensible way."""
        sub = df.select_dtypes(include=["datetime64", "timedelta64"])
        out = {}
        for col in sub.columns:
            out[col] = _format_time_column(df[col])
        return df.assign(**out)

    def _get_filename(path_ser, strip_extension):
        """Get the file name from a path series."""
        ser = path_ser.astype(str)
        split_ser = ser.str.split("/")
        if strip_extension:
            file_names = [x[-1].split(".")[0] for x in split_ser]
        else:
            file_names = [x[-1] for x in split_ser]
        return pd.Series(file_names)

    # Validate inputs. Note we cannot use the validation decorator or
    # it introduces a circular import.
    prefix = TypeAdapter(str).validate_python(prefix)
    attrs = TypeAdapter(tuple[str, ...]).validate_python(attrs)
    coords = TypeAdapter(tuple[str, ...]).validate_python(coords)
    sep = TypeAdapter(str).validate_python(sep)

    # Ensure we are working with a dataframe.
    df = dc.scan_to_df(
        patch_data,
        exclude=(),
    )
    if df.empty:
        return pd.Series(dtype=str)
    col_set = set(df.columns)
    # Handle special cases.
    if "name" in col_set:
        return df["name"].astype(str)
    if "path" in col_set:
        return _get_filename(df["path"], strip_extension)
    # Determine the requested fields and get the ones that are there.
    coord_fields = zip([f"{x}_min" for x in coords], [f"{x}_max" for x in coords])
    requested_fields = list(attrs) + list(*coord_fields)
    current = set(df.columns)
    fields = [x for x in requested_fields if x in current]
    # Get a sub dataframe and convert any datetime things to strings.
    sub = df[fields].pipe(_format_time_columns).fillna("").astype(str)
    out = f"{prefix}_{sep}" + sub[fields[0]].str.cat(sub[fields[1:]], sep=sep)
    return out


def get_dim_axis_value(
    patch: PatchType,
    *,
    args: tuple = tuple(),
    kwargs: dict = FrozenDict(),
    arg_keys: tuple[str] = ("dim", "coord", "dims", "coords"),
    allow_multiple: bool = False,
    allow_extra: bool = False,
) -> tuple[_DimAxisValue, ...]:
    """
    Get dimension nane, index, and values from args/kwargs for a patch.

    This is helpful for implementing flexible fetching of dimension name,
    corresponding patch axis, and function specific values from args and
    kwargs as inputs.

    Parameters
    ----------
    patch
        The patch which contains desired dimensions.
    args
        A tuple of possible dimension names.
    kwargs
        A dict of dimension_name: value
    arg_keys
        Keys in the dictionary that indicate
    allow_multiple
        If True, allow multiple dimensions to be selected.
    allow_extra
        If True, do not raise an error if extra args or kwargs are found.

    Returns
    -------
    Returns a tuple of:
        ((dim, axis, value), (dim, axis, value), ...)
    To support retreiving multiple values from the same inputs. If dim name
    is found in args, its corresponding values is `None`.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.patch import get_dim_axis_value
    >>>
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Get tuple of dimension name, axis, and value from dict (eg kwargs)
    >>> (dim, ax, val) = get_dim_axis_value(patch, kwargs={"time": 10})[0]
    >>> assert dim == "time" and ax == patch.dims.index("time") and val == 10
    >>>
    >>> # Get dim name and axis from tuple (eg args)
    >>> (dim, ax, val) = get_dim_axis_value(patch, args=("time",))[0]
    >>> assert dim == "time" and ax == patch.dims.index("time") and val is None
    >>>
    >>> # Get list of dim, ax val from multiple kwargs and args
    >>> info = get_dim_axis_value(
    ...     patch, args=("time", ), kwargs={"distance": 10}, allow_multiple=True,
    ... )
    >>> assert len(info) == 2
    """
    kwargs = dict(kwargs)  # copy kwargs to avoid modifying the input dict
    dims: tuple[str, ...] = patch.dims
    # Pop out any args implicit in kwargs.
    args = args + tuple(kwargs.pop(x) for x in arg_keys if x in kwargs)
    input_set = set(args) | set(kwargs)
    patch_dim_set = set(dims)
    overlap = patch_dim_set & input_set
    # Determine if there is the right number of overlaps.
    if not overlap or (len(overlap) > 1 and not allow_multiple):
        msg = (
            "You must use exactly one dimension name in args or kwargs. "
            f"You passed the following kwargs: {kwargs} args: {args } "
            f"to a patch with dimensions {patch.dims}"
        )
        raise PatchCoordinateError(msg)
    # Handle the case of extra inputs
    if (remaining := input_set - patch_dim_set) and not allow_extra:
        msg = (
            "The following input dimensions are not found in the patch. " f"{remaining}"
        )
        raise PatchCoordinateError(msg)
    # Ensure order is preserved (eg args, then kwargs)
    dim_out = tuple(x for x in args + tuple(kwargs) if x in overlap)
    # Package everything up and return
    out = tuple(_DimAxisValue(x, dims.index(x), kwargs.get(x)) for x in dim_out)
    return out


def get_dim_sampling_rate(patch: PatchType, dim: str) -> float:
    """
    Get sampling rate, as a float from sampling period along a dimension.

    Parameters
    ----------
    patch
        The input patch.
    dim
        Dimension to extract.

    Raises
    ------
    [CoordDataError](`dascore.exceptions.CoordDataError`) if patch is not
    evenly sampled along desired dimension.
    """
    d_dim = patch.coords.coord_map[dim].step
    if isinstance(d_dim, np.timedelta64):
        d_dim = d_dim / np.timedelta64(1, "s")
    if pd.isnull(d_dim):
        # get the name of the calling function
        calling_function = inspect.getframeinfo(sys._getframe(1))[2]
        msg = (
            f"Patch coordinate {dim} is not evenly sampled as required by "
            f"{calling_function}. This can be fixed with Patch.snap or "
            f"Patch.extrapolate. "
        )
        raise CoordDataError(msg)
    return 1.0 / d_dim


def _get_data_units_from_dims(patch, dims, operator):
    """Get new data units from some operation on dimensions."""
    if (data_units := get_quantity(patch.attrs.data_units)) is None:
        return
    dim_units = None
    for dim in iterate(dims):
        dim_unit = get_quantity(patch.get_coord(dim).units)
        if dim_unit is None:
            continue
        dim_units = dim_unit if dim_units is None else dim_unit * dim_units
    if dim_units is not None:
        data_units = operator(data_units, dim_units)
    return data_units


def _get_dx_or_spacing_and_axes(
    patch,
    dim,
    require_sorted=True,
    require_evenly_spaced=False,
) -> tuple[tuple[float | np.ndarray, ...], tuple[int, ...]]:
    """
    Return dx (spacing) or values for a list of dims and corresponding axes.

    Parameters
    ----------
    patch
        The input patch
    dim
        The dimension name or sequence of such
    require_sorted
        If True, raise an error if all requested dimensions are not sorted.
    require_evenly_spaced
        If True, raise an error if all requested dimensions are not evenly sampled.
    """
    dims = iterate(dim if dim is not None else patch.dims)
    out = []
    axes = []
    for dim_ in dims:
        coord = patch.get_coord(
            dim_,
            require_sorted=require_sorted,
            require_evenly_sampled=require_evenly_spaced,
        )
        if coord.evenly_sampled:
            val = coord.step
        else:
            val = coord.data
        # need to convert val to float so datetimes work
        out.append(to_float(val))
        axes.append(patch.dims.index(dim_))

    return tuple(out), tuple(axes)


def align_patch_coords(
    patch1: PatchType, patch2: PatchType
) -> tuple[PatchType, PatchType]:
    """
    Align patches so they have compatible attrs and data broadcast together.

    Parameters
    ----------
    patch1
        The first patch.
    patch2
        The second patch.
    """
    # Fast path for no alignment needed
    if patch1.coords == patch2.coords:
        return patch1, patch2
    shared_dims = set(patch1.dims) & set(patch2.dims)
    if not shared_dims:
        msg = (
            "Cannot align patches with no shared dimensions. Dimensions are "
            f"patch1: {patch1.dims}, patch2: {patch2.dims}"
        )
        raise PatchCoordinateError(msg)
    # First ensure the patches have the same dims
    dims = _merge_tuples(patch1.dims, patch2.dims)
    dim_dict = {x: num for num, x in enumerate(dims)}
    patch1 = patch1.append_dims(*dims).transpose(*dims)
    patch2 = patch2.append_dims(*dims).transpose(*dims)
    # Next, find the common coordinates and align.
    align_1, align_2 = [slice(None)] * len(dims), [slice(None)] * len(dims)
    new_coords_1, new_coords_2 = {}, {}
    for dim in shared_dims:
        coord1, coord2 = patch1.get_coord(dim), patch2.get_coord(dim)
        if coord1 == coord2:
            continue
        dim_ind = dim_dict[dim]
        # We actually need to do some alignment here.
        ncoord1, ncoord2, sli1, sli2 = coord1.align_to(coord2)
        new_coords_1[dim], new_coords_2[dim] = ncoord1, ncoord2
        align_1[dim_ind], align_2[dim_ind] = sli1, sli2
    # No alignment needed, just skip.
    if not (new_coords_1 or new_coords_2):
        return patch1, patch2
    # Update coordinate managers and reshape arrays, return new patches.
    coord1 = patch1.coords.update(**new_coords_1)
    coord2 = patch2.coords.update(**new_coords_2)
    array1 = _apply_union_indexers(tuple(align_1), patch1.data)
    array2 = _apply_union_indexers(tuple(align_2), patch2.data)
    out1 = patch1.new(data=array1, coords=coord1)
    out2 = patch2.new(data=array2, coords=coord2)
    return out1, out2


def check_dims(
    patch1,
    patch2,
    check_behavior: WARN_LEVELS = "raise",
    intersection: bool = False,
) -> bool:
    """
    Return True if dimensions of two patches are equal.

    Parameters
    ----------
    patch1
        first patch
    patch2
        second patch
    check_behavior
        String with 'raise' will raise an error if incompatible,
        'warn' will provide a warning, None will do nothing.
    intersection
        If True, allow any intersection of dimensions to pass. This is useful
        when only broad-castablity needs to be checked. If false require dims
        to be equal.
    """
    dims1, dims2 = patch1.dims, patch2.dims
    dims_ok = True
    if not intersection and patch1.dims == patch2.dims:
        return True
    dset1, dset2 = set(dims1), set(dims2)
    if intersection and (dset1 | dset2):
        return True
    msg = (
        "Patch dimensions are not compatible for merging."
        f" Patch1 dims: {dims1}, Patch2 dims: {dims2}"
    )
    warn_or_raise(msg, exception=IncompatiblePatchError, behavior=check_behavior)
    return dims_ok


def check_coords(
    patch1,
    patch2,
    check_behavior: WARN_LEVELS = "raise",
    dim_to_ignore=None,
    ignore_dim_eq_shape=True,
) -> bool:
    """
    Return True if the coordinates of two patches are compatible, else False.

    Parameters
    ----------
    patch1
        patch 1
    patch2
        patch 2
    check_behavior
        String with 'raise' will raise an error if incompatible,
        'warn' will provide a warning.
    dim_to_ignore
        None by default (all coordinates must be identical).
        String specifying a dimension that differences in values,
        but not shape, are allowed.
    ignore_dim_eq_shape
        If True, the ignored dims must be equal shape to pass check.
        If dim_to_ignore is None this has no effect.
    """
    cm1 = patch1.coords
    cm2 = patch2.coords
    cset1, cset2 = set(cm1.coord_map), set(cm2.coord_map)
    shared = cset1 & cset2
    not_equal_coords = []
    for coord in shared:
        coord1 = cm1.coord_map[coord]
        coord2 = cm2.coord_map[coord]
        if coord1 == coord2:
            # Straightforward case, coords are identical.
            continue
        elif coord == dim_to_ignore:
            # If dimension that's ok to ignore value differences,
            # check whether shape is the same.
            if coord1.shape == coord2.shape:
                continue
            elif ignore_dim_eq_shape:
                not_equal_coords.append(coord)
        else:
            not_equal_coords.append(coord)
    if not_equal_coords and len(shared):
        msg = (
            f"Patches are not compatible. The following shared coordinates "
            f"are not equal: {coord}"
        )
        warn_or_raise(msg, exception=IncompatiblePatchError, behavior=check_behavior)
        return False
    return True


def _merge_aligned_coords(cm1, cm2):
    """Merge aligned coordinates removing non coords."""
    assert cm1.dims == cm2.dims, "dimensions are not aligned"
    out = {}
    for name in set(cm1.coord_map) & set(cm2.coord_map):
        coord1 = cm1.coord_map[name]
        coord2 = cm2.coord_map[name]
        dim1, dim2 = cm1.dim_map.get(name), cm2.dim_map.get(name)
        # Coords already equal, just use first.
        if coord1.approx_equal(coord2) and dim1 == dim2:
            out[name] = (dim1, coord1)
        # Deal with Non coords
        non_count = sum([coord1._partial, coord2._partial])
        if non_count == 1:
            out[name] = (dim1, coord1 if coord2._partial else coord2)
        elif non_count == 2:
            out[name] = (dim1, coord1 if coord1.size > coord2.size else coord2)
        assert name in out
    return cm1.update(**out)


def _merge_models(attrs1, attrs2, coord=None, attrs_to_ignore=DEFAULT_ATTRS_TO_IGNORE):
    """Ensure models are equal in the right ways, merge together."""
    no_comp_keys = set(attrs_to_ignore)
    if attrs1 == attrs2:
        return attrs1
    dict1, dict2 = dict(attrs1), dict(attrs2)
    if coord is not None:
        new_coords = coord.to_summary_dict()
        dict1["coords"], dict2["coords"] = new_coords, new_coords
    else:
        dict1.pop("coords"), dict2.pop("coords")
    common_keys = set(dict1) & set(dict2)
    ne_attrs = []
    for key in common_keys:
        if key in no_comp_keys:
            continue
        if dict2[key] != dict1[key]:
            ne_attrs.append(key)
    if ne_attrs:
        msg = (
            "Patches are not compatible because the following attributes "
            f"are not equal. {ne_attrs}"
        )
        raise IncompatiblePatchError(msg)
    return combine_patch_attrs([dict1, dict2], conflicts="keep_first")


def merge_compatible_coords_attrs(
    patch1: PatchType,
    patch2: PatchType,
    attrs_to_ignore=DEFAULT_ATTRS_TO_IGNORE,
    dim_intersection: bool = False,
    validate_coords: bool = True,
) -> tuple[dc.core.CoordManager, dc.PatchAttrs]:
    """
    Merge the coordinates and attributes of patches or raise if incompatible.

    The rules for compatibility are:

    - All attrs must be equal other than those specified in attrs_to_ignore.
    - Patches must share the same dimensions unless dim_intersection == True.
    - All shared dimensional coordinates must be strictly equal
    - If patches share a non-dimensional coordinate they must be equal.

    Any coordinates or attributes contained by a single patch will be included
    in the output.

    Parameters
    ----------
    patch1
        The first patch
    patch2
        The second patch
    attrs_to_ignore
        A sequence of attributes to not consider in equality. Only these
        attributes from the first patch are kept in outputs.
    dim_intersection
        If True, merge if any dimensions overlap, else raise if all do not
        overlap.
    validate_coords
        If True, ensure the coords are equal, else the responsibility for this
        was handled upstream.
    """

    def _merge_coords(coords1, coords2):
        out = {}
        cmap1, cmap2 = coords1.coord_map, coords2.coord_map
        coord_names = set(cmap1) | set(cmap2)
        # fast path to update identical coordinates
        if coord_names == set(cmap1):
            return coords1
        if coord_names == set(cmap2):
            return coords2
        # otherwise just squish coords from both managers together.
        for name in coord_names:
            coord = coords1 if name in coords1.coord_map else coords2
            dims = coord.dim_map[name]
            out[name] = (dims, coord.coord_map[name])
        # Need to get coordinate that are in output, but preserve order.
        dims = _merge_tuples(coords1.dims, coords2.dims)
        return dc.core.coordmanager.get_coord_manager(out, dims=dims)

    check_dims(patch1, patch2, intersection=dim_intersection)
    if validate_coords:
        check_coords(patch1, patch2)
    coord1, coord2 = patch1.coords, patch2.coords
    attrs1, attrs2 = patch1.attrs, patch2.attrs
    coord_out = _merge_coords(coord1, coord2)
    attrs = _merge_models(attrs1, attrs2, coord_out, attrs_to_ignore=attrs_to_ignore)
    return coord_out, attrs


def _add_history_str(attrs, hist_str):
    """Add a single string to the history attribute in attrs."""
    new_history = list(attrs.history)
    new_history.append(hist_str)
    return attrs.update(history=new_history)


def _spool_up(func):
    """
    Spool the output of a function.

    This is primarily to turn methods that return a list of patches
    into something that can be used as a spool method.
    """

    @functools.wraps(func)
    def _wrapper(self, *args, **kwargs):
        """Wrapper for function."""
        out = func(self, *args, **kwargs)
        return dc.spool(out)

    return _wrapper


@compose_docstring(check_bev=check_behavior_description)
def concatenate_patches(
    patches: Sequence[dc.Patch] | dc.BaseSpool,
    check_behavior: WARN_LEVELS = "warn",
    **kwargs,
) -> Sequence[dc.Patch]:
    """
    Concatenate the patches together.

    Only patches which are compatible with the first patch are concatenated
    together.

    Parameters
    ----------
    {check_bev}
    **kwargs
        Used to specify the dimension and number of patches to merge
        together. A value of None attempts to concatenate all patches
        into as single patch.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Concatenate patches along time axis
    >>> spool = dc.spool([patch, patch])
    >>> spool_concat = spool.concatenate(time=None)
    >>> assert len(spool_concat) == 1
    >>>
    >>> # Concatenate patches along a new dimension.
    >>> # Note: This will only include the first patch if existing
    >>> # dimensions are not identical.
    >>> spool_concat = spool.concatenate(wave_rank=None)
    >>> assert "wave_rank" in spool_concat[0].dims
    >>>
    >>> # Concatenate patches in groups of 3.
    >>> big_spool = dc.spool([patch] * 12)
    >>> spool_concat = big_spool.concatenate(time=3)
    >>> assert len(spool_concat) == 4

    Notes
    -----
    - [`Spool.chunk`](`dascore.BaseSpool.chunk`) performs a similar operation
      but accounts for coordinate values.
    - See also the
      [chunk section of the spool tutorial](`docs/tutorial/spool`#concatenate)
    """

    def _get_dim_and_value(kwargs):
        """Get the dimension name and value"""
        if not len(kwargs) == 1:
            msg = "Exactly one keyword argument must be passed to concatenate."
            raise ParameterError(msg)
        assert len(kwargs) == 1
        [(dim, val)] = kwargs.items()

        return dim, val

    def get_compatible_patches(patches, dim, check_behavior):
        """Get the patches which can be concatenated, dim names, and new dim."""
        # We need to drop private coords for dft concats to work.
        patches = list(x.drop_private_coords() for x in patches)
        first_patch = patches[0]
        compat_patches = []
        # Ensure patch dimensions are compatible.
        dim_set = {x.dims for x in patches}
        if not len(dim_set) == 1:
            msg = "Cannot concatenate patches with different dimensions."
            raise PatchCoordinateError(msg)
        # Get dim name and such
        first_dims = next(iter(dim_set))
        new_dim = dim not in first_dims
        dims = tuple([*list(first_dims), dim]) if new_dim else first_dims
        # Get patches compatible with first.
        for p in patches:
            dims_ok = check_dims(first_patch, p, check_behavior)
            coords_ok = check_coords(
                patch1=first_patch,
                patch2=p,
                check_behavior=check_behavior,
                dim_to_ignore=dim,
                ignore_dim_eq_shape=False,
            )
            if dims_ok and coords_ok:
                compat_patches.append(p)
        return compat_patches, dims, new_dim

    def get_output_array(patches, axis, new_dim):
        """Get a list of output arrays."""
        sub_arrays = [x.data[..., None] if new_dim else x.data for x in patches]
        out = np.concatenate(sub_arrays, axis=axis)
        return out

    def _get_new_coords(patch_list, dim, new_dim):
        """Get new coordinates for creating patch."""
        coords = patch_list[0].coords
        if new_dim:
            coords = coords.update(**{dim: (dim, len(patch_list))})
        else:
            array_list = [x.get_array(dim) for x in patch_list]
            array = np.concatenate(array_list, axis=0)
            coords = coords.update(**{dim: array})
        return coords

    dim, val = _get_dim_and_value(kwargs)
    patches, dims, new_dim = get_compatible_patches(patches, dim, check_behavior)
    out = []
    for patch_list in yield_sub_sequences(patches, val):
        ar = get_output_array(patch_list, dims.index(dim), new_dim)
        attrs = _add_history_str(patch_list[0].attrs, "concatenate")
        coords = _get_new_coords(patch_list, dim, new_dim)
        out.append(dc.Patch(data=ar, attrs=attrs, coords=coords, dims=dims))
    return out


@compose_docstring(check_desc=check_behavior_description)
def stack_patches(
    patches, dim_vary=None, check_behavior: WARN_LEVELS = "warn"
) -> PatchType:
    """
    Stack (add) all patches compatible with first patch together.

    Parameters
    ----------
    dim_vary
        The name of the dimension which can be different in values
        (but not shape) and patches still added together.
        If None, all dimension values must be equal.
    {check_desc}

    Examples
    --------
    >>> import dascore as dc
    >>> # add a spool with equal sized patches but progressing time dim
    >>> spool = dc.get_example_spool()
    >>> stacked_patch = spool.stack(dim_vary='time')
    """
    # check the dims/coords of first patch (considered to be standard for rest)
    init_patch = patches[0]
    stack_arr = np.zeros_like(init_patch.data)

    # ensure dim_vary is in dims
    if dim_vary is not None and dim_vary not in init_patch.dims:
        msg = f"Dimension {dim_vary} is not in first patch."
        raise PatchCoordinateError(msg)

    for p in patches:
        # check dimensions of patch compared to init_patch
        dims_ok = check_dims(init_patch, p, check_behavior)
        coords_ok = check_coords(init_patch, p, check_behavior, dim_vary)
        # actually do the stacking of data
        if dims_ok and coords_ok:
            stack_arr = stack_arr + p.data

    # create attributes for the stack with adjusted history
    stack_attrs = _add_history_str(init_patch.attrs, "stack")

    # create coords array for the stack
    stack_coords = init_patch.coords
    if dim_vary:  # adjust dim_vary to start at 0 for junk dimension indicator
        coord_to_change = stack_coords.coord_map[dim_vary]
        new_dim = coord_to_change.update_limits(min=0)
        stack_coords = stack_coords.update_coords(**{dim_vary: new_dim})
    return dc.Patch(stack_arr, stack_coords, init_patch.dims, stack_attrs)



"""Utilities for segy."""

from __future__ import annotations

import datetime
import warnings

import numpy as np

# --- Getting format/version
import pandas as pd

import dascore as dc
from dascore import to_float
from dascore.core import get_coord_manager
from dascore.exceptions import InvalidSpoolError, PatchError
from dascore.utils.misc import optional_import


def twos_comp(bytes_):
    """Get twos complement of bytestring."""
    bits = len(bytes_) * 8
    val = int.from_bytes(bytes_, "big")
    if (val & (1 << (bits - 1))) != 0:  # if sign bit is set e.g., 8bit: 128-255
        val = val - (1 << bits)  # compute negative value
    return val  # return positive value as is


def _get_segy_version(fp):
    """
    Determine if file handle contains segy data.

    Returns (segy, version) if so else False.

    Based on ObsPy's implementation writen by Lion Krischer.
    https://github.com/obspy/obspy/blob/master/obspy/io/segy/core.py
    """
    # Read 400byte header into byte string.
    fp.seek(3200)
    header = fp.read(400)
    data_trace_count = twos_comp(header[12:14])
    auxiliary_trace_count = twos_comp(header[14:16])
    sample_interval = twos_comp(header[16:18])
    samples_per_trace = twos_comp(header[20:22])
    data_format_code = twos_comp(header[24:26])
    format_number_major = twos_comp(header[300:301])
    format_number_minor = twos_comp(header[301:302])
    fixed_len_flag = twos_comp(header[302:304])

    checks = (
        # First check that some samples are defined.
        samples_per_trace > 0,
        # Then ensure the sample intervals is defined. This can be defined in trace
        # header so 0 is ok, but not negative numbers.
        sample_interval >= 0,
        # Ensure the data sample format code is valid using range in 2,1 standard
        1 <= data_format_code <= 16,
        # Check version code
        format_number_major in {0, 1, 2, 3},
        # Sanity checks for other values.
        data_trace_count >= 0,
        auxiliary_trace_count >= 0,
        format_number_minor in {0, 1, 2, 3, 100},
        fixed_len_flag in {0, 1},
    )
    if all(checks):
        version = f"{format_number_major}.{format_number_minor}"
        return "segy", version
    else:
        return False


def _get_filtered_data_and_coords(segy_fi, coords, time=None, channel=None):
    """
    Read data from segy_file, possibly reading only subsections.

    Return filtered data and update coord manager.
    """
    traces_inds_to_read = np.arange(len(segy_fi.header), dtype=np.int64)
    time_slice = slice(None, None)
    traces = segy_fi.trace

    # filter time
    if time is not None:
        time_coord = coords.coord_map["time"]
        new_coord, time_slice = time_coord.select(time)
        coords = coords.update(time=new_coord)

    # filter channel
    if channel:
        channel_coord = coords.coord_map["channel"]
        new_coord, channel_inds = channel_coord.select(channel)
        coords = coords.update(channel=new_coord)
        traces_inds_to_read = traces_inds_to_read[channel_inds]

    # filter channels
    data_list = [traces[x][time_slice] for x in traces_inds_to_read]
    return np.stack(data_list, axis=-1), coords


def _get_coords(fi):
    """
    Get coordinates of the segy file.

    Time comes from the SEGY format of year, julian day, hour, minute, second.

    Distance axis is channel number. If the user knows the delta_x,
    then the axis should be modified.

    If a user knows the dx, change from channel to distance using
    patch.update_coords after reading
    """
    segyio = optional_import("segyio")
    trace_field = segyio.TraceField
    header_0 = fi.header[0]

    # Get time array from SEGY headers
    starttime = _get_time_from_header(header_0)
    dt = dc.to_timedelta64(header_0[trace_field.TRACE_SAMPLE_INTERVAL] / 1_000_000)
    ns = header_0[trace_field.TRACE_SAMPLE_COUNT]
    time_array = starttime + dt * np.arange(ns)

    # Get distance array from SEGY header
    channel = np.arange(len(fi.header))

    coords = get_coord_manager(
        {"time": time_array, "channel": channel}, dims=("time", "channel")
    )
    return coords


def _get_attrs(fi, coords, path, file_io):
    """Create Patch Attribute from SEGY header contents."""
    attrs = dc.PatchAttrs(
        path=path,
        file_version=file_io.version,
        file_format=file_io.name,
        coords=coords,
    )
    return attrs


def _get_time_from_header(header):
    """Creates a datetime64 object from SEGY header date information."""
    segyio = optional_import("segyio")
    trace_field = segyio.TraceField

    year = header[trace_field.YearDataRecorded]
    julday = header[trace_field.DayOfYear]
    hour = header[trace_field.HourOfDay]
    minute = header[trace_field.MinuteOfHour]
    second = header[trace_field.SecondOfMinute]
    # make those timedate64
    fmt = "%Y.%j.%H.%M.%S"
    s = f"{year}.{julday}.{hour}.{minute}.{second}"
    time = datetime.datetime.strptime(s, fmt)
    return dc.to_datetime64(time)


def _get_patch_with_channel_coord(patch):
    """Ensure the patch has a channel coordinate."""
    dims = set(patch.dims)
    non_time = next(iter(dims - {"time"}))
    msg = (
        "Currently the segy writer only handles 'channel' as the non-time "
        "dimension this results in a loss of the '{non_time}' dimension."
    )
    warnings.warn(msg)
    coord = patch.get_coord(non_time)
    array = np.arange(len(coord))
    patch = patch.update_coords(**{non_time: array}).rename_coords(
        **{non_time: "channel"}
    )
    return patch


def _get_segy_compatible_patch(spool, round_error_max=3e-9):
    """
    Get a patch that will be writable as a segy file.
    Ensure coords are ("channel", "time").
    """
    # Ensure we have a single patch with coordinates time and distance.
    spool = [spool] if isinstance(spool, dc.Patch) else spool
    if len(spool) != 1:
        msg = "Can only write a spool with as single patch as segy."
        raise InvalidSpoolError(msg)
    patch = spool[0]
    dims = set(patch.dims)
    has_distance_or_channel = dims & {"distance", "channel"}
    if len(dims) != 2 or "time" not in dims or not has_distance_or_channel:
        msg = (
            "Can only save 2D patches to SEGY with a time dimension and "
            "either channel or distance dimensions."
        )
        raise PatchError(msg)
    # Currently we only support channels not distance dimension.
    if "channel" not in dims:
        patch = _get_patch_with_channel_coord(patch)
    # Ensure there will be no loss in the time sampling.
    # segy supports us precision
    time_step = dc.to_float(patch.get_coord("time").step)
    new_samp = np.round(time_step, 6)
    round_error = np.abs(new_samp - time_step).max()
    if round_error > round_error_max:
        msg = (
            f"The segy format support us precision for temporal sampling. "
            f"The input patch has a time step of {time_step} which will result "
            "in a loss of precision. Either manually set the time step with "
            "patch.update_coords or resample the time axis with patch.resample"
        )
        raise PatchError(msg)
    return patch.transpose("channel", "time")


def _make_time_header_dict(time_coord):
    """Make the time header dict from a time coordinate."""
    header = {}
    timestamp = pd.Timestamp(dc.to_datetime64(time_coord.min()))
    time_step_ms = np.round(to_float(time_coord.step) * 1_000_000)

    segyio = optional_import("segyio")
    trace_field = segyio.TraceField

    header[trace_field.YearDataRecorded] = timestamp.year
    header[trace_field.DayOfYear] = timestamp.day_of_year
    header[trace_field.HourOfDay] = timestamp.hour
    header[trace_field.MinuteOfHour] = timestamp.minute
    header[trace_field.SecondOfMinute] = timestamp.second
    header[trace_field.TRACE_SAMPLE_INTERVAL] = int(time_step_ms)
    header[trace_field.TRACE_SAMPLE_COUNT] = len(time_coord)

    return header


def _write_segy(spool, resource, version, segyio):
    """
    Private function for writing a patch/spool as SEGY.
    """
    patch = _get_segy_compatible_patch(spool)
    time, channel = patch.get_coord("time"), patch.get_coord("channel")
    chanel_step = channel.step

    time_dict = _make_time_header_dict(time)
    bin_field = segyio.BinField
    spec = segyio.spec()

    spec.format = 1  # 1 means float32 TODO look into supporting more
    spec.samples = np.ones(len(time)) * len(channel)
    spec.ilines = range(len(channel))
    spec.xlines = [1]

    # For 32 bit float for now.
    data = patch.data.astype(np.float32)

    with segyio.create(resource, spec) as f:
        # Update the file header info.
        f.bin.update(tsort=segyio.TraceSortingFormat.INLINE_SORTING)
        f.bin.update(
            {
                bin_field.Samples: time_dict[segyio.TraceField.TRACE_SAMPLE_COUNT],
                bin_field.Interval: time_dict[segyio.TraceField.TRACE_SAMPLE_INTERVAL],
                bin_field.SEGYRevision: int(version.split(".")[0]),
                bin_field.SEGYRevisionMinor: int(version.split(".")[1]),
            }
        )
        # Then iterate each channel and dump to segy.
        for num, data in enumerate(data):
            header = dict(time_dict)
            header.update(
                {
                    segyio.su.offset: chanel_step,
                    segyio.su.iline: num,
                    segyio.su.xline: 1,
                }
            )
            f.header[num] = header
            f.trace[num] = data



"""
Compatibility module for DASCore.

All components/functions that may be exchanged for other numpy/scipy
compatible libraries should go in this model.
"""

from __future__ import annotations

from contextlib import suppress

import numpy as np
from numpy import floor, interp  # NOQA
from numpy.random import RandomState
from rich.progress import Progress  # NOQA
from scipy.interpolate import interp1d  # NOQA
from scipy.ndimage import zoom  # NOQA
from scipy.signal import decimate, resample, resample_poly  # NOQA

random_state = RandomState(42)


class DataArray:
    """A dummy class for when xarray isn't installed."""


with suppress(ImportError):
    from xarray import DataArray  # NOQA


def array(array):
    """Wrapper function for creating 'immutable' arrays."""
    out = np.asarray(array)
    # Setting the write flag to false makes the array immutable unless
    # the flag is switched back.
    out.setflags(write=False)
    return out


def is_array(maybe_array):
    """Determine if an object is array like."""
    # This is here so that we can support other array types in the future.
    return isinstance(maybe_array, np.ndarray)



"""
Module for static, matplotlib-based visualizations and figure generation.
"""
from __future__ import annotations
from dascore.utils.misc import MethodNameSpace

from .spectrogram import spectrogram
from .waterfall import waterfall
from .wiggle import wiggle
from .map_fiber import map_fiber


class VizPatchNameSpace(MethodNameSpace):
    """A class for storing visualization namespace."""

    waterfall = waterfall
    spectrogram = spectrogram
    wiggle = wiggle
    map_fiber = map_fiber



"""Module for performing differentiation on patches."""

from __future__ import annotations

from collections.abc import Sequence
from operator import truediv

import numpy as np

from dascore.compat import is_array
from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.utils.misc import broadcast_for_index, iterate, optional_import
from dascore.utils.patch import (
    _get_data_units_from_dims,
    _get_dx_or_spacing_and_axes,
    patch_function,
)


def _grad_diff(ar, axes, dxs):
    """Perform differentiation on the data using np.grad."""
    # iteratively apply diff along each axis.
    for ax, dx in zip(axes, dxs):
        ar = np.gradient(ar, dx, axis=ax, edge_order=2)
    return ar


def _findiff_diff(data, axis, order, dx):
    """Use findiff to get differences along specified dimensions."""
    # If more than one axis, need to make tuples.
    findiff = optional_import("findiff")
    if len(axis) > 1:
        orders = [1] * len(axis)  # always first order (first derivative).
        tuples = [tuple(x) for x in zip(axis, dx, orders)]
        func = findiff.FinDiff(*tuples, acc=order)
    else:
        func = findiff.FinDiff(axis[0], dx[0], 1, acc=order)
    out = func(data)
    return out


def _get_diff(order, data, axes, dx_or_spacing):
    """Get the diff of an array along specified axes."""
    if order == 2:
        new_data = _grad_diff(data, axes, dx_or_spacing)
    else:
        new_data = _findiff_diff(data, axes, order, dx_or_spacing)
    return new_data


def _strided_diff(order, patch, axes, dx_or_spacing, step):
    """Calculate a strided differentiation along specified axes."""
    if len(axes) > 1:
        msg = "Step in patch.differentiate can only be used along one axis."
        raise ParameterError(msg)
    new_data = np.empty_like(patch.data)
    dx_or_space = dx_or_spacing[0]
    for step_ in range(step):
        current_slice = slice(step_, None, step)
        slicer = broadcast_for_index(patch.ndim, axes[0], current_slice)
        # Need to either double DX or slice the coordinate spacing to
        # account for the fact we are skipping some columns/rows.
        if is_array(dx_or_space):
            _dx_or_space = dx_or_space[current_slice]
        else:
            _dx_or_space = dx_or_space * step
        sub = _get_diff(order, patch.data[slicer], axes, [_dx_or_space])
        new_data[slicer] = sub
    return new_data


@patch_function()
def differentiate(
    patch: PatchType,
    dim: str | Sequence[str] | None,
    order: int = 2,
    step: int = 1,
) -> PatchType:
    r"""
    Calculate first derivative along dimension(s) using centeral diferences.

    The shape of the output patch is the same as the input patch. Derivative
    along edges are calculated with the same order (accuarcy) as centeral
    points using non-centered stencils.

    Parameters
    ----------
    patch
        The patch to differentiate.
    dim
        The dimension(s) along which to differentiate. If None differentiates
        over all dimensions.
    order
        The order of the differentiation operator. Must be a possitive, even
        integar.
    step
        The number of columns/rows to skip for differention.
        eg: an array of [a b c d e] uses b and d to calculate diff of c when
        step = 1 and order = 2. When step = 2, a and e are used to calcualte
        diff at c.

    Notes
    -----
    For order=2 (the default) numpy's gradient function is used. When
    order != 2, the optional package findiff must be installed in which case
    order is interpreted as accuracy ("order" means order of differention
    in that package).

    The second order first derivate, for an evenly spaced coordiante,
    is defined as:

    $$
    \hat{f}(x) = \frac{f(x + dx) - f(x - dx)}{2dx} + O({dx}^2)
    $$

    Where $\hat{f}(x)$ is the estiamted derivative of $f$ at $x$, $dx$ is
    the sample spacing, and $O$ is the error term.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Example 1
    >>> # 2nd order stencil for 1st derivative along time dimension
    >>> patch_diff_1 = patch.differentiate(dim='time', order=2)
    >>>
    >>> # Example 2
    >>> # 1st derivative along all dimensions
    >>> patch_diff_2 = patch.differentiate(dim=None)
    >>>
    >>> # Example 3
    >>> # 1st derivative using a step size of 3. This spaces out the columns
    >>> # or rows used for estimating the derivative.
    >>> patch_diff_3 = patch.differentiate(dim="distance", step=3, order=2)
    """
    dims = iterate(dim if dim is not None else patch.dims)
    dx_or_spacing, axes = _get_dx_or_spacing_and_axes(patch, dims)
    if step > 1:
        new_data = _strided_diff(order, patch, axes, dx_or_spacing, step)
    # This avoids an extra copy of the array so probably merits its own case.
    else:
        new_data = _get_diff(order, patch.data, axes, dx_or_spacing)
    # update units
    data_units = _get_data_units_from_dims(patch, dims, truediv)
    attrs = patch.attrs.update(data_units=data_units)
    return patch.new(data=new_data, attrs=attrs)



"""Tests for waterfall plots."""

from __future__ import annotations

import matplotlib.pyplot as plt
import pytest

import dascore as dc
from dascore.utils.time import is_datetime64


def check_label_units(patch, ax, dims, color="distance"):
    """Ensure patch label units match axis."""
    axis_dict = {0: "xaxis", 1: "yaxis"}
    # dims = []
    # Check coord-inate names
    for coord_name in dims:
        coord = patch.coords.coord_map[coord_name]
        if is_datetime64(coord[0]):
            continue  # just skip datetimes for now.
        index = dims.index(coord_name)
        axis = getattr(ax, axis_dict[index])
        label_text = axis.get_label().get_text().lower()
        assert str(coord.units.units) in label_text
        assert coord_name in label_text
    # check colorbar labels
    # cax = ax.images[-1].colorbar
    coord = patch.coords.coord_map[color]
    yaxis_label = ax.figure.get_children()[-1].yaxis.label.get_text()
    assert str(coord.units.units) in yaxis_label


@pytest.fixture(scope="session")
def patch_random_start(event_patch_1):
    """Get a patch with a random, odd, starttime."""
    random_starttime = dc.to_datetime64("2020-01-02T02:12:11.02232")
    attrs = dict(event_patch_1.attrs)
    coords = {i: v for i, v in event_patch_1.coords.items()}
    time = coords["time"] - coords["time"].min()
    coords["time"] = time + random_starttime
    attrs["time_min"] = coords["time"].min()
    attrs["time_max"] = coords["time"].max()
    patch = event_patch_1.update(attrs=attrs, coords=coords)
    return patch


class TestPlotMap:
    """Tests for map plot."""

    def test_str_input(self, random_patch_with_lat_lon):
        """Call map_fiber plot, return."""
        patch = random_patch_with_lat_lon.set_units(latitude="ft")
        patch = patch.set_units(longitude="m")
        ax = patch.viz.map_fiber("latitude", "longitude")

        caxis_label = ax.figure.get_children()[-1].yaxis.label.get_text()

        # check labels
        assert "latitude" in ax.get_xlabel().lower()
        assert "longitude" in ax.get_ylabel().lower()
        assert "distance" in caxis_label
        assert isinstance(ax, plt.Axes)

    def test_array_inputs(self, random_patch_with_lat_lon):
        """Call map_fiber plot, return."""
        lats = random_patch_with_lat_lon.coords.get_array("latitude")
        lons = random_patch_with_lat_lon.coords.get_array("longitude")
        data = 0.5 * (lats + lons)
        ax = random_patch_with_lat_lon.viz.map_fiber(lats, lons, data)

        assert isinstance(ax, plt.Axes)

    def test_default_parameters(self, random_patch):
        """Call map_fiber plot, return."""
        ax = random_patch.viz.map_fiber()

        # check labels
        assert "distance" in ax.get_ylabel().lower()
        assert "distance" in ax.get_xlabel().lower()
        assert isinstance(ax, plt.Axes)

    def test_colorbar_scale(self, random_patch):
        """Tests for the scaling parameter."""
        ax_scalar = random_patch.viz.map_fiber(scale=0.2)
        assert ax_scalar is not None
        seq_scalar = random_patch.viz.map_fiber(scale=[0.1, 0.3])
        assert seq_scalar is not None

    def test_colorbar_absolute_scale(self, random_patch):
        """Tests for absolute scaling of colorbar."""
        patch = random_patch.new(data=random_patch.data * 100 - 50)
        ax1 = patch.viz.map_fiber(scale_type="absolute", scale=(-50, 50))
        assert ax1 is not None
        ax2 = patch.viz.map_fiber(scale_type="absolute", scale=10)
        assert ax2 is not None

    def test_no_colorbar(self, random_patch):
        """Ensure the colorbar can be disabled."""
        ax = random_patch.viz.map_fiber(cmap=None)
        # ensure no colorbar was created.
        assert len(ax.figure.get_children()) == 2

    def test_units(self, random_patch_with_lat_lon):
        """Test that units show up in labels."""
        # standard units

        pa = random_patch_with_lat_lon.set_units(distance="m/s")
        ax = pa.viz.map_fiber()
        check_label_units(pa, ax, ["distance", "distance"])

        new = pa.set_units(latitude="ft", longitude="m")
        ax = new.viz.map_fiber("latitude", "longitude")
        check_label_units(new, ax, ["latitude", "longitude"])

    def test_show(self, random_patch, monkeypatch):
        """Ensure show path is callable."""
        monkeypatch.setattr(plt, "show", lambda: None)
        random_patch.viz.map_fiber(show=True)



"""Tests for waterfall plots."""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import pytest

import dascore as dc
from dascore.units import get_quantity_str
from dascore.utils.time import is_datetime64, to_timedelta64


def check_label_units(patch, ax):
    """Ensure patch label units match axis."""
    axis_dict = {0: "yaxis", 1: "xaxis"}
    dims = patch.dims
    # Check coord-inate names
    for coord_name in dims:
        coord = patch.coords.coord_map[coord_name]
        if is_datetime64(coord[0]):
            continue  # just skip datetimes for now.
        index = dims.index(coord_name)
        axis = getattr(ax, axis_dict[index])
        label_text = axis.get_label().get_text().lower()
        assert str(coord.units.units) in label_text
        assert coord_name in label_text
    # check colorbar labels
    cax = ax.images[-1].colorbar
    yaxis_label = cax.ax.yaxis.label.get_text()
    assert str(patch.attrs.data_units.units) in yaxis_label
    assert str(patch.attrs.data_type) in yaxis_label


@pytest.fixture(scope="session")
def patch_random_start(event_patch_1):
    """Get a patch with a random, odd, starttime."""
    random_starttime = dc.to_datetime64("2020-01-02T02:12:11.02232")
    attrs = dict(event_patch_1.attrs)
    coords = {i: v for i, v in event_patch_1.coords.items()}
    time = coords["time"] - coords["time"].min()
    coords["time"] = time + random_starttime
    attrs["time_min"] = coords["time"].min()
    attrs["time_max"] = coords["time"].max()
    patch = event_patch_1.update(attrs=attrs, coords=coords)
    return patch


@pytest.fixture(scope="session")
def patch_two_times(random_patch):
    """Create a patch with two time dims."""
    dist = random_patch.coords.get_array("distance")
    pa = random_patch.update_coords(distance=dc.to_datetime64(dist))
    return pa


class TestWaterfall:
    """Tests for waterfall plot."""

    @pytest.fixture()
    def timedelta_patch(self, random_patch):
        """Make a patch with one dimension dtype of timedelta64."""
        old_coord = random_patch.get_coord("time")
        new_time = to_timedelta64(np.arange(len(old_coord)))
        return random_patch.update_coords(time=new_time)

    def test_returns_axes(self, random_patch):
        """Call waterfall plot, return."""
        # modify patch to include line at start
        data = np.array(random_patch.data)
        data[:100, :100] = 2.0  # create an origin block for testing axis line up
        data[:100, -100:] = -2.0  #
        out = random_patch.new(data=data)
        ax = out.viz.waterfall()

        # check labels
        assert random_patch.dims[0] in ax.get_ylabel().lower()
        assert random_patch.dims[1] in ax.get_xlabel().lower()
        assert isinstance(ax, plt.Axes)

    def test_colorbar_scale(self, random_patch):
        """Tests for the scaling parameter."""
        ax_scalar = random_patch.viz.waterfall(scale=0.2)
        assert ax_scalar is not None
        seq_scalar = random_patch.viz.waterfall(scale=[0.1, 0.3])
        assert seq_scalar is not None

    def test_colorbar_absolute_scale(self, random_patch):
        """Tests for absolute scaling of colorbar."""
        patch = random_patch.new(data=random_patch.data * 100 - 50)
        ax1 = patch.viz.waterfall(scale_type="absolute", scale=(-50, 50))
        assert ax1 is not None
        ax2 = patch.viz.waterfall(scale_type="absolute", scale=10)
        assert ax2 is not None

    def test_doc_intro_example(self, event_patch_1):
        """Simple test to ensure the doc examples can be run."""
        patch = event_patch_1.pass_filter(time=(None, 300))
        _ = patch.viz.waterfall(scale=0.04)
        _ = patch.transpose("distance", "time").viz.waterfall(scale=0.04)

    def test_time_axis_label_int_overflow(self, random_patch):
        """Make sure the time axis labels are correct (windows compatibility)."""
        ax = random_patch.viz.waterfall()
        name = ["y", "x"][random_patch.dims.index("time")]
        # Get the piece of the label corresponding to the starttime
        # WE can just grab the offset text.
        sub_ax = getattr(ax, f"{name}axis")
        plt.tight_layout()  # need to call this to get offset to show up.
        offset_str = sub_ax.get_major_formatter().get_offset()
        min_time = random_patch.coords.get_array("time").min()
        assert str(min_time).startswith(offset_str)

    def test_no_colorbar(self, random_patch):
        """Ensure the colorbar can be disabled."""
        ax = random_patch.viz.waterfall(cmap=None)
        # ensure no colorbar was created.
        assert ax.images[-1].colorbar is None

    def test_units(self, random_patch):
        """Test that units show up in labels."""
        # standard units
        pa = random_patch.set_units("m/s")
        ax = pa.viz.waterfall()
        check_label_units(pa, ax)
        # weird units
        new = pa.set_units(
            "furlongs/fortnight",
            distance="feet",
        )
        ax = new.viz.waterfall()
        check_label_units(new, ax)

    def test_time_no_units(self, patch_two_times):
        """time-like dims shouldn't show units in label."""
        pa = patch_two_times
        dims = pa.dims
        ax = pa.viz.waterfall()
        assert ax.get_xlabel() == dims[1]
        assert ax.get_ylabel() == dims[0]

    def test_patch_with_data_type(self, random_patch):
        """Ensure a patch with data_type titles the colorbar."""
        patch = random_patch.update_attrs(
            data_type="strain rate",
            data_units="1/s",
        )
        ax = patch.viz.waterfall()
        check_label_units(patch, ax)

    def test_timedelta_axis(self, timedelta_patch):
        """Ensure plot works when one axis has timedleta dtype. See #309."""
        # if this doesnt raise it probably works ;)
        ax = timedelta_patch.viz.waterfall()
        assert ax is not None

    def test_show(self, random_patch, monkeypatch):
        """Ensure show path is callable."""
        monkeypatch.setattr(plt, "show", lambda: None)
        random_patch.viz.waterfall(show=True)

    def test_log(self, random_patch):
        """Ensure log is callable."""
        ax = random_patch.viz.waterfall(log=True)

        # Retrieve the colorbar label
        cb = ax.get_figure().get_axes()[-1]
        cb_label = cb.get_ylabel()

        # Retrieve the expected data type and data units
        data_type = str(random_patch.attrs["data_type"])
        data_units = get_quantity_str(random_patch.attrs.data_units) or ""
        expected_dunits = f" ({data_units})" if data_units else ""

        # Construct the expected label
        expected_label = f"{data_type}{expected_dunits} - log_10"

        # Check if the colorbar label matches the expected label
        assert (
            cb_label == expected_label
        ), f"Expected '{expected_label}', but got '{cb_label}'"



"""Tests for Dispersion transforms."""

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import ParameterError
from dascore.transform.taup import _jit_taup_general, _jit_taup_uniform, tau_p
from dascore.utils.misc import suppress_warnings
from dascore.utils.time import to_timedelta64


def linear_slope_patch(nch, nt, vel, t0, dist=[]):
    """Returns the a patch with linear slope event"""
    # Create attributes, or metadata
    attrs = dict(category="DAS", id="linear_slope", data_units="um/(m * s)")

    dt = 1 / 1000

    if np.size(dist) != nch:
        distance_start = 0
        dx = 1.0
        distance_step = dx
        distance = distance_start + np.arange(nch) * distance_step
    else:
        distance = dist

    # Create coordinates, labels for each axis in the array.
    time_start = dc.to_datetime64("2017-09-18")
    time_step = to_timedelta64(dt)
    time = time_start + np.arange(nt) * time_step

    coords = dict(time=time, distance=distance)

    # Define dimensions (first label corresponds to data axis 0)
    dims = ("distance", "time")

    array = np.zeros(shape=(nch, nt))

    if vel > 0:
        for i in range(nch):
            t = int((t0 + distance[i] / vel) / dt)
            array[i, t] += 1.0
    else:
        for i in range(nch):
            t = int((t0 + np.abs((distance[-1] - distance[i]) / vel)) / dt)
            array[i, t] += 1.0

    pa = dc.Patch(data=array, coords=coords, attrs=attrs, dims=dims)
    return pa


class TestTauP:
    """Tests for the tau-p module."""

    @pytest.fixture(scope="class")
    def tau_p_patch(self, random_patch):
        """Returns the random patched transformed to tau-p."""
        pytest.importorskip("numba")
        test_vels = np.linspace(1500, 5000, 100)

        with suppress_warnings(DeprecationWarning):
            out = tau_p(random_patch, test_vels)
        return out

    def test_tau_p_consistency(self, tau_p_patch):
        """Checks consistency of tau_p module."""
        pytest.importorskip("numba")
        # assert time dimension
        assert "time" in tau_p_patch.dims
        # assert slowness dimension
        assert "slowness" in tau_p_patch.dims

        p_vals = tau_p_patch.coords.get_array("slowness")
        n_p_vals = len(p_vals)

        assert n_p_vals % 2 == 0
        n_p_vals = int(n_p_vals / 2)

        assert np.array_equal(
            p_vals[0:n_p_vals], -1.0 / np.flip(np.linspace(1500, 5000, 100))
        )
        assert np.array_equal(
            p_vals[n_p_vals : 2 * n_p_vals], 1.0 / np.linspace(1500, 5000, 100)
        )

    def test_negative_vel_raises(self, random_patch):
        """Ensures negative velocities raise an error"""
        msg = "Input velocities must be positive."
        velocities = np.array([1000, 1100, -200, 1000])
        with pytest.raises(ParameterError, match=msg):
            random_patch.tau_p(velocities=velocities)

    def test_non_monotonic_vel_raises(self, random_patch):
        """Ensures non monotonic velocities raise an error"""
        msg = "Input velocities must be monotonically increasing."
        velocities = np.array([1000, 1100, 300, 1200])
        with pytest.raises(ParameterError, match=msg):
            random_patch.tau_p(velocities=velocities)

        velocities = np.array([1000, 1100, 1100, 1200])
        with pytest.raises(ParameterError, match=msg):
            random_patch.tau_p(velocities=velocities)

    def test_slowness_vals(self):
        """Ensures correct slowness and tau values are computed"""
        pytest.importorskip("numba")
        test_vels = np.linspace(1000, 3000, 101)
        nch = 1000
        nt = 2000

        # positive slope
        vel = 1500
        t0 = 0.3
        linear_patch = linear_slope_patch(nch, nt, vel, t0)
        tau_p_patch = linear_patch.tau_p(test_vels)
        p_vals = tau_p_patch.get_coord("slowness")
        t_vals = tau_p_patch.get_coord("time")
        a = tau_p_patch.data
        (p_ind, t_ind) = np.unravel_index(a.argmax(), a.shape)
        assert np.abs(1.0 / p_vals[p_ind] - vel) < 20
        assert np.abs((t_vals[t_ind] - t_vals[0]) / np.timedelta64(1, "s") - t0) < 0.02

        # negative slope
        vel = -2000
        t0 = 0.5
        linear_patch = linear_slope_patch(nch, nt, vel, t0)
        tau_p_patch = linear_patch.tau_p(test_vels)
        p_vals = tau_p_patch.get_coord("slowness")
        t_vals = tau_p_patch.get_coord("time")
        a = tau_p_patch.data
        (p_ind, t_ind) = np.unravel_index(a.argmax(), a.shape)
        assert np.abs(1.0 / p_vals[p_ind] - vel) < 20
        assert np.abs((t_vals[t_ind] - t_vals[0]) / np.timedelta64(1, "s") - t0) < 0.02

        dist = np.zeros(nch)
        cumdist = 0
        for i in range(nch):
            dist[i] = cumdist
            cumdist = cumdist + 1.0 + 1.0 * (i % 2 == 0)

        # positive slope, non-equal distance
        vel = 1800
        t0 = 0.4
        linear_patch = linear_slope_patch(nch, nt, vel, t0, dist)
        tau_p_patch = linear_patch.tau_p(test_vels)
        p_vals = tau_p_patch.get_coord("slowness")
        t_vals = tau_p_patch.get_coord("time")
        a = tau_p_patch.data
        (p_ind, t_ind) = np.unravel_index(a.argmax(), a.shape)
        assert np.abs(1.0 / p_vals[p_ind] - vel) < 20
        assert np.abs((t_vals[t_ind] - t_vals[0]) / np.timedelta64(1, "s") - t0) < 0.02

        # negative slope, non-equal distance
        vel = -1700
        t0 = 0.1
        linear_patch = linear_slope_patch(nch, nt, vel, t0, dist)
        tau_p_patch = linear_patch.tau_p(test_vels)
        p_vals = tau_p_patch.get_coord("slowness")
        t_vals = tau_p_patch.get_coord("time")
        a = tau_p_patch.data
        (p_ind, t_ind) = np.unravel_index(a.argmax(), a.shape)
        assert np.abs(1.0 / p_vals[p_ind] - vel) < 20
        assert np.abs((t_vals[t_ind] - t_vals[0]) / np.timedelta64(1, "s") - t0) < 0.02

    def test_coverage(self):
        """
        A small test of the jit'ed functions in python mode for coverage reasons.
        """
        pytest.importorskip("numba")
        data = np.zeros((10, 10))
        dx = 1
        dt = 1
        p_vals = np.array([1])
        _jit_taup_uniform.func(data, dx, dt, p_vals)
        distance = np.arange(10) * dt
        _jit_taup_general.func(data, distance, dt, p_vals)



"""IO module for reading Sentek's DAS data format."""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.io import BinaryReader
from dascore.io.core import FiberIO

from .utils import _get_patch_attrs, _get_version


class SentekV5(FiberIO):
    """Support for Sentek Instrument data format."""

    name = "sentek"
    version = "5"
    preferred_extensions = ("das",)

    def read(
        self,
        resource: BinaryReader,
        time=None,
        distance=None,
        **kwargs,
    ) -> dc.BaseSpool:
        """Read a Sentek das file, return a DataArray."""
        attrs, coords, offsets = _get_patch_attrs(resource)
        resource.seek(offsets[0])
        array = np.fromfile(resource, dtype=np.float32, count=offsets[1] * offsets[2])
        array = np.reshape(array, (offsets[1], offsets[2])).T
        patch = dc.Patch(data=array, attrs=attrs, coords=coords, dims=coords.dims)
        # Note: we are being a bit sloppy here in that selecting on time/distance
        # doesn't actually affect how much data is read from the binary file. This
        # is probably ok though since Sentek files tend to be quite small.
        return dc.spool(patch).select(time=time, distance=distance)

    def get_format(self, resource: BinaryReader, **kwargs) -> tuple[str, str] | bool:
        """Auto detect sentek format."""
        return _get_version(resource)

    def scan(self, resource: BinaryReader, **kwargs):
        """Extract metadata from sentek file."""
        extras = {
            "file_format": self.name,
            "file_version": self.version,
            "path": resource.name,
        }

        return [_get_patch_attrs(resource, extras=extras)[0]]



"""A deprecated module."""

from __future__ import annotations

import warnings

from dascore.core import attrs


def __getattr__(name):
    out = getattr(attrs, name)
    msg = (
        "The DASCore.core.schema module is deprecated. Use DAScore.core.attrs "
        "instead."
    )
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    return out



"""Tests for RSF format."""

import os
from pathlib import Path

import numpy as np
import pytest

import dascore as dc
from dascore.io.rsf import RSFV1


class TestRsfWrite:
    """testing the rSF write out function."""

    def test_write_nopath(self, random_patch, tmp_path):
        """
        Test write function with no binary path specified.
        Data and header are combined.
        """
        spool = dc.spool(random_patch)
        path = tmp_path / "test_hdrdata.rsf"
        RSFV1().write(spool, path)

        assert path.exists()
        test_data = random_patch.data.astype(np.float32)
        dtype = np.dtype(test_data.dtype)
        file_esize = dtype.itemsize
        datasize = test_data.size * file_esize
        assert os.path.getsize(path) >= datasize

    def test_write_path(self, random_patch, tmp_path):
        """Test write function with different binary data path specified."""
        spool = dc.spool(random_patch)
        path = tmp_path / "test_hdr.rsf"
        datapath = tmp_path / "binary/test_data.rsf"
        RSFV1().write(spool, path, data_path=datapath)

        assert path.exists()
        newdatapath = Path(str(datapath) + "@")
        assert newdatapath.exists()
        test_data = random_patch.data.astype(np.float32)
        dtype = np.dtype(test_data.dtype)
        file_esize = dtype.itemsize
        datasize = test_data.size * file_esize
        assert os.path.getsize(newdatapath) == datasize

    def test_write_cmplx(self, random_patch, tmp_path, **kwargs):
        """Test write function for non-int and non-float values.
        if should fail and return:
        ValueError("Data format is not integer or floating.").
        """
        complex_patch = random_patch.dft("time")
        spool = dc.spool(complex_patch)
        path = tmp_path / "test_hdrcmplx.rsf"
        with pytest.raises(ValueError):
            RSFV1().write(spool, path)

    def test_write_int(self, random_patch, tmp_path, **kwargs):
        """Test write function for int values.
        if should return float values of the integer numbers.
        """
        data = np.ones_like(random_patch.data, dtype=np.int32)
        int_patch = random_patch.new(data=data)
        spool = dc.spool(int_patch)
        path = tmp_path / "test_hdrint.rsf"
        datapath = tmp_path / "binary/test_int.rsf"
        RSFV1().write(spool, path, data_path=datapath)

        assert path.exists()
        newdatapath = Path(str(datapath) + "@")
        assert newdatapath.exists()
        test_data = data.astype(np.float32)
        dtype = np.dtype(test_data.dtype)
        file_esize = dtype.itemsize
        datasize = test_data.size * file_esize
        assert os.path.getsize(newdatapath) == datasize



"""Processing for applying a taper."""

from __future__ import annotations

from collections.abc import Sequence
from functools import reduce
from operator import add

import numpy as np
from scipy.signal import windows  # the best operating system?

from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.units import Quantity
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import broadcast_for_index
from dascore.utils.patch import get_dim_axis_value, patch_function
from dascore.utils.time import to_float

TAPER_FUNCTIONS = dict(
    barthann=windows.barthann,
    bartlett=windows.bartlett,
    blackman=windows.blackman,
    blackmanharris=windows.blackmanharris,
    bohman=windows.bohman,
    hamming=windows.hamming,
    hann=windows.hann,
    cos=windows.hann,
    nuttall=windows.nuttall,
    parzen=windows.parzen,
    triang=windows.triang,
    ramp=windows.triang,
)


def _get_taper_slices(patch, kwargs):
    """Get slice for start/end of patch."""
    dim, axis, value = get_dim_axis_value(patch, kwargs=kwargs)[0]
    coord = patch.coords.coord_map[dim]
    if isinstance(value, Sequence | np.ndarray):
        assert len(value) == 2, "Length 2 sequence required."
        start, stop = value[0], value[1]
    else:
        start, stop = value, value
    dur = coord.max() - coord.min()
    # either let units pass through or multiply by d_len
    clses = (Quantity, np.timedelta64)
    start = start if isinstance(start, clses) or start is None else start * dur
    stop = stop if isinstance(stop, clses) or stop is None else stop * dur
    stop = -stop if stop is not None else stop
    _, inds_1 = coord.select((None, start), relative=True)
    _, inds_2 = coord.select((stop, None), relative=True)
    return axis, (start, stop), inds_1, inds_2


def _get_window_function(window_type):
    """Get the window function to use for taper."""
    # get taper function or raise if it isn't known.
    if window_type not in TAPER_FUNCTIONS:
        msg = (
            f"'{window_type}' is not a known window type. "
            f"Options are: {sorted(TAPER_FUNCTIONS)}"
        )
        raise ParameterError(msg)
    func = TAPER_FUNCTIONS[window_type]
    return func


def _validate_windows(samps, start_slice, end_slice, shape, axis):
    """Validate the windows don't overlap or exceed dim len."""
    max_len = shape[axis]
    start_ind = start_slice.stop
    end_ind = end_slice.start

    bad_start = samps[0] is not None and (start_ind is None or start_ind < 0)
    bad_end = samps[1] is not None and (end_ind is None or end_ind > max_len)

    if bad_start or bad_end:
        msg = "Total taper lengths exceed total dim length"
        raise ParameterError(msg)

    if start_ind is None or end_ind is None:
        return
    if start_ind > end_ind:
        msg = "Taper windows cannot overlap"
        raise ParameterError(msg)


@patch_function()
@compose_docstring(taper_type=sorted(TAPER_FUNCTIONS))
def taper(
    patch: PatchType,
    window_type: str = "hann",
    **kwargs,
) -> PatchType:
    """
    Taper the ends of the signal.

    Parameters
    ----------
    patch
        The patch instance.
    window_type
        The type of window to use For tapering. Supported Options are:
            {taper_type}.
    **kwargs
        Used to specify the dimension along which to taper and the percentage
        of total length of the dimension or absolute units. If a single value
        is passed, the taper will be applied to both ends. A length two tuple
        can specify different values for each end, or no taper on one end.

    Returns
    -------
    The tapered patch.

    See Also
    --------
    [Patch.taper_range](`dascore.Patch.taper_range`)

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch() # generate example patch
    >>>
    >>> # Apply an Hanning taper to 5% of each end for time dimension.
    >>> patch_taper1 = patch.taper(time=0.05, window_type="hann")
    >>>
    >>> # Apply a triangular taper to 10% of the start of the distance dimension.
    >>> patch_taper2 = patch.taper(distance=(0.10, None), window_type='triang')
    >>>
    >>> # Apply taper on first and last 15 m along distance axis.
    >>> from dascore.units import m
    >>> patch_taper3 = patch.taper(distance=15 * m)
    """
    func = _get_window_function(window_type)
    # get taper values in samples.
    out = np.array(patch.data)  # Need to make a copy here.
    shape = out.shape
    n_dims = len(out.shape)
    axis, samps, start_slice, end_slice = _get_taper_slices(patch, kwargs)
    _validate_windows(samps, start_slice, end_slice, shape, axis)
    if samps[0] is not None:
        val = start_slice.stop
        window = func(2 * val)[:val]
        # get indices window (which will broadcast) and data
        data_inds = broadcast_for_index(n_dims, axis, start_slice)
        window_inds = broadcast_for_index(n_dims, axis, slice(None), fill=None)
        out[data_inds] = out[data_inds] * window[window_inds]
    if samps[1] is not None:
        val = shape[axis] - end_slice.start
        window = func(2 * val)[val:]
        data_inds = broadcast_for_index(n_dims, axis, end_slice)
        window_inds = broadcast_for_index(n_dims, axis, slice(None), fill=None)
        out[data_inds] = out[data_inds] * window[window_inds]
    return patch.new(data=out)


def _get_taper_coord_inds(coord, values, relative, samples):
    """Get the index of the referenced coord inds."""
    error_msg = "A len 2 or 4 sequence is required for taper values"
    if not isinstance(values, (Sequence | np.ndarray)) or not len(values):
        raise ParameterError(error_msg)
    # More than 1 sequence was passed, recurse to flatten out.
    elif isinstance(values[0], (Sequence | np.ndarray)):
        out = [_get_taper_coord_inds(coord, x, relative, samples) for x in values]
        return reduce(add, out)
    elif len(values) not in {2, 4}:
        raise ParameterError(error_msg)
    # Ok inputs, convert to index along coordinate.
    out = [None] * len(values)
    for num, val in enumerate(values):
        if val is None or val == ...:
            if len(values) == 2:
                msg = "Cannot use ... or None when only two values provided"
                raise ParameterError(msg)
            # None or ... means min_val in first half of list else max_val
            out[num] = 0 if (num / len(out)) < 0.5 else len(coord) - 1
        else:
            out[num] = coord.get_next_index(val, samples=samples, relative=relative)
    # Always need a len 4 sequence
    if len(out) == 2:
        out = [0, *out, len(coord)]
    return [out]  # return a list of len4 sequences.


def _get_taper_curve(coord, ind_1, ind_2, window_type, reverse=False):
    """Get the taper curve between index1 and index2."""
    func = _get_window_function(window_type)
    samps = ind_2 - ind_1
    taper = func(samps * 2 + 1)[:samps]
    if reverse:
        taper = taper[::-1]
    # Need to extrapolate to get correct values for non evenly sampled coords.
    if not coord.evenly_sampled:
        # The current window represents the snapped (evenly sampled) coords
        old_coord = coord.select((ind_1, ind_2), samples=True)[0]
        new_coord = old_coord.snap().change_length(len(old_coord))
        old_x, new_x = to_float(old_coord.values), to_float(new_coord.values)
        assert len(old_x) == len(new_x) == len(taper)
        taper = np.interp(new_x, old_x, taper)
    return taper


def _get_range_envelope(coord, inds, window_type, invert):
    """Create a broadcast envelope for taper."""
    out = np.zeros(len(coord))
    for ind_set in inds:
        assert len(ind_set) == 4
        i1, i2, i3, i4 = ind_set
        left_taper = _get_taper_curve(coord, i1, i2, window_type)
        right_taper = _get_taper_curve(coord, i3, i4, window_type, reverse=True)
        out[i1:i2] += left_taper
        out[i3:i4] += right_taper
        out[i2:i3] += 1
    if invert:
        out = np.abs(out - np.max(out))

    return out


@patch_function()
@compose_docstring(taper_type=sorted(TAPER_FUNCTIONS))
def taper_range(
    patch: PatchType,
    window_type: str = "hann",
    invert=False,
    relative=False,
    samples=False,
    **kwargs,
) -> PatchType:
    """
    Taper a range inside the patch.

    Parameters
    ----------
    patch
        A patch instance.
    window_type
        The type of window to use For tapering. Supported Options are:
            {taper_type}.
    invert
        If True, the values inside the specified range are set to zero
        and gradually tapered to 1.
    samples
        If True, the values specified by the kwargs indicate samples
        rather than values along the indicated dimension.
    relative
        If True, the values specified in kwargs are relateive to the
        start (if positive) or end (if negative) of the indicated
        dimension.
    **kwargs
        Used to specify the dimension along which to taper. Values can be
        either a length 2 sequence or a length 4 sequence. If len == 2
        then the left taper starts at [0] and ends at the start of
        the coordinate. The left taper starts at [1] and ends at
        the end of the coordinate. If len == 4, values between [1] and [2]
        are left alone, values between [0] and [1] as well as values between
        [2] and [3] are gradually tapered. Values outside of this range are
        set to 0.

    Returns
    -------
    The tapered patch.

    See Also
    --------
    [Patch.taper](`dascore.Patch.taper`)

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>>
    >>> epatch = dc.get_example_patch()
    >>> patch = epatch.new(data=np.ones_like(epatch.data))
    >>>
    >>> # Taper values outside of specified times to zero
    >>> t1 = dc.to_datetime64("2017-09-18T00:00:04")
    >>> t2 = dc.to_datetime64("2017-09-18T00:00:07")
    >>> patch_tapered_1 = patch.taper_range(time=(t1, t2))
    >>>
    >>> # Taper values inside specified range to 0
    >>> patch_tapered_2 = patch.taper_range(time=(t1, t2), invert=True)
    >>>
    >>> # Specify taper range (4 values) such that values outside
    >>> # that range are 0, between [0] and [1] as well as
    >>> # [2] and [3] are tapered and values inside [1] and [2] are
    >>> # not effected.
    >>> patch_tapered_3 = patch.taper_range(
    ...     time=(1, 2, 5, 5),
    ...     relative=True,
    ... )
    >>>
    >>> # Use samples rather than absolute time values.
    >>> patch_tapered_4 = patch.taper_range(
    ...     distance=(10, 80),
    ...     samples=True
    ... )
    >>>
    >>> # Apply two non-overlapping tapers
    >>> taper_range = ((25,50,100,125), (150,175,200,225))
    >>> patch_tapered_5 = patch.taper_range(distance=taper_range)

    """
    dim, ax, values = get_dim_axis_value(patch, kwargs=kwargs)[0]
    coord = patch.get_coord(dim, require_sorted=True)
    inds = _get_taper_coord_inds(coord, values, relative, samples)
    env = _get_range_envelope(coord, inds, window_type, invert)
    # Ensure envelope broadcasts to index
    indexer = broadcast_for_index(patch.ndim, ax, value=slice(None), fill=None)
    env_broadcastable = env[indexer]
    return patch.new(data=patch.data * env_broadcastable)



"""Misc. tests for Terra15."""

from __future__ import annotations

import shutil

import numpy as np
import pandas as pd
import pytest
import tables

import dascore as dc
from dascore.io.terra15 import Terra15FormatterV4


class TestTerra15:
    """Misc tests for Terra15."""

    @pytest.fixture(scope="class")
    def missing_gps_terra15_hdf5(self, terra15_v5_path, tmp_path_factory):
        """Creates a terra15 file with missing GPS Time."""
        new = tmp_path_factory.mktemp("missing_gps") / "missing.hdf5"
        shutil.copy(terra15_v5_path, new)
        with tables.open_file(new, "a") as fi:
            fi.root.data_product.gps_time._f_remove()
        return new

    def test_missing_gps_time(self, missing_gps_terra15_hdf5):
        """Tests for when GPS time isn't found."""
        patch = dc.read(missing_gps_terra15_hdf5)[0]
        assert isinstance(patch, dc.Patch)
        assert not np.any(pd.isnull(patch.coords.get_array("time")))

    def test_time_slice(self, terra15_v6_path):
        """Ensure time slice within the file works."""
        info = dc.scan_to_df(terra15_v6_path).iloc[0]
        file_t1, file_t2 = info["time_min"], info["time_max"]
        dur = file_t2 - file_t1
        new_dur = dur / 4
        t1, t2 = file_t1 + new_dur, file_t1 + 2 * new_dur
        out = dc.read(terra15_v6_path, time=(t1, t2))[0]
        assert isinstance(out, dc.Patch)
        attrs = out.attrs
        assert attrs.time_min >= t1
        assert attrs.time_max <= t2

    def test_time_slice_no_snap(self, terra15_v6_path):
        """Ensure no snapping returns raw time."""
        info = dc.scan_to_df(terra15_v6_path).iloc[0]
        file_t1, file_t2 = info["time_min"], info["time_max"]
        dur = file_t2 - file_t1
        new_dur = dur / 4
        t1, t2 = file_t1 + new_dur, file_t1 + 2 * new_dur
        out = dc.read(terra15_v6_path, time=(t1, t2), snap_dims=False)[0]
        assert isinstance(out, dc.Patch)
        attrs = out.attrs
        assert attrs.time_min >= t1
        assert attrs.time_max <= t2

    def test_units(self, terra15_das_patch):
        """All units should be defined on terra15 patch."""
        patch = terra15_das_patch
        assert patch.attrs.data_units is not None
        assert patch.attrs.distance_units is not None
        assert patch.attrs.time_units is not None
        assert patch.get_coord("time").units == patch.attrs.time_units
        assert patch.get_coord("distance").units == patch.attrs.distance_units

    def test_hdf5file_not_terra15(self, generic_hdf5):
        """Assert that the generic hdf5 file is not a terra15."""
        parser = Terra15FormatterV4()
        assert not parser.get_format(generic_hdf5)


class TestTerra15Unfinished:
    """Test for reading files with zeroes filled at the end."""

    @pytest.fixture(scope="class")
    def patch_unfinished(self, terra15_das_unfinished_path):
        """Return the patch with zeroes at the end."""
        out = dc.spool(terra15_das_unfinished_path)[0]
        return out

    def test_zeros_gone(self, patch_unfinished):
        """No zeros should exist in the data."""
        data = patch_unfinished.data
        all_zero_rows = np.all(data == 0, axis=1)
        assert not np.any(all_zero_rows)

    def test_monotonic_time(self, patch_unfinished):
        """Ensure the time is increasing."""
        time = patch_unfinished.coords.get_array("time")
        assert np.all(np.diff(time) >= np.timedelta64(0, "s"))



"""Tau-p Patch transforms."""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.units import convert_units
from dascore.utils.jit import maybe_numba_jit
from dascore.utils.patch import patch_function
from dascore.utils.time import to_float


@maybe_numba_jit(nopython=True)
def _jit_taup_uniform(data, dx, dt, p_vals):
    """
    A numba version of uniform Tau-P transform.

    See [`tau_p`](`dascore.transform.taup.tau_p`) for details.
    """
    (nx, nt) = data.shape
    n_slo = p_vals.size
    two_sided_p_vals = np.concatenate((-np.flip(p_vals), p_vals))
    taup = np.zeros(shape=(2 * n_slo, nt))

    for ip in numba.prange(n_slo):  # noqa
        pdx = p_vals[ip] * dx
        for tau in range(nt):
            for ix in range(nx):
                samp_val = tau + pdx * ix / dt
                ind = int(samp_val)
                if ind + 1 < nt:
                    taup[ip + n_slo, tau] += (1.0 + ind - samp_val) * data[ix, ind] + (
                        samp_val - ind
                    ) * data[ix, ind + 1]
                    taup[n_slo - 1 - ip, tau] += (1.0 + ind - samp_val) * data[
                        nx - 1 - ix, ind
                    ] + (samp_val - ind) * data[nx - 1 - ix, ind + 1]
    return two_sided_p_vals, taup


@maybe_numba_jit(nopython=True)
def _jit_taup_general(data, distance, dt, p_vals):
    """
    A numba version of general Tau-P transform.

    See [`tau_p`](`dascore.transform.taup.tau_p`) for details.
    """
    (nx, nt) = data.shape
    n_slo = p_vals.size
    two_sided_p_vals = np.concatenate((-np.flip(p_vals), p_vals))
    taup = np.zeros(shape=(2 * n_slo, nt))
    mod_distance = distance - distance[0]

    for ip in numba.prange(n_slo):  # noqa
        p = p_vals[ip]
        for tau in range(nt):
            for ix in range(nx):
                samp_val_pos = tau + p * mod_distance[ix] / dt
                ind_pos = int(samp_val_pos)
                samp_val_neg = (
                    tau + p * (mod_distance[-1] - mod_distance[nx - 1 - ix]) / dt
                )
                ind_neg = int(samp_val_pos)
                if ind_pos + 1 < nt:
                    taup[ip + n_slo, tau] += (1.0 + ind_pos - samp_val_pos) * data[
                        ix, ind_pos
                    ] + (samp_val_pos - ind_pos) * data[ix, ind_pos + 1]
                if ind_neg + 1 < nt:
                    taup[n_slo - 1 - ip, tau] += (1.0 + ind_neg - samp_val_neg) * data[
                        nx - 1 - ix, ind_neg
                    ] + (samp_val_neg - ind_neg) * data[nx - 1 - ix, ind_neg + 1]
    return two_sided_p_vals, taup


@patch_function(required_dims=("time", "distance"))
def tau_p(
    patch: PatchType,
    velocities: NDArray[np.floating],
) -> PatchType:
    """
    Compute linear tau-p transform.

    Parameters
    ----------
    patch
        Patch to transform. Has to have dimensions of time and distance.
    velocities
        NumPY array of velocities, in m/s if units are not attached,
        for which to compute slowness (p).

    Notes
    -----
    - Output will always be double the size of vels, with negative velocities
      (right-to-left) first, followed by positive velocities (left-to-right).

    - Uses linear interpolation in time

    Example
    -------
    ```{python}
    >>> import dascore as dc
    >>> import numpy as np
    >>>
    >>> patch = (
    ...    dc.get_example_patch('example_event_1')
    ... )

    >>> taup_patch = (
    ...     patch.taper(time=0.1)
    ...     .pass_filter(time=(..., 300))
    ...     .tau_p(np.arange(1000,6000,10))
    ...     .transpose('time','slowness')
    ... )
    >>> ax = taup_patch.viz.waterfall(show=False, cmap=None)
    >>> _ = taup_patch.viz.waterfall(ax=ax)
    """
    patch_cop = patch.convert_units(distance="m", time="s").transpose(
        "distance", "time"
    )
    dist = patch_cop.get_coord("distance")
    time = patch_cop.get_coord("time", require_evenly_sampled=True)
    dt = to_float(time.step)

    if np.any(velocities <= 0):
        msg = "Input velocities must be positive."
        raise ParameterError(msg)

    if not np.all(np.diff(velocities) > 0):
        raise ParameterError("Input velocities must be monotonically increasing.")

    # Handle unit conversions if needed.
    velocities = convert_units(velocities, to_units="m/s")

    # Chooses code version based on whether distance between channels
    # is uniform or not
    if dist.evenly_sampled:
        func = _jit_taup_uniform
        dist_val = dist.step
    else:
        func = _jit_taup_general
        dist_val = dist.values

    slowness, tau_p_data = func(patch.data, dist_val, dt, 1.0 / velocities)

    attrs = patch.attrs.update(category="taup")
    coords = dict(slowness=slowness, time=time)

    tau_p_patch = patch.new(
        data=tau_p_data, coords=coords, attrs=attrs, dims=["slowness", "time"]
    )
    return tau_p_patch.set_units(slowness="s/m", time="s")



"""
Module for reading data stored in xml_binary format.

This format is a directory which contains metadata as a single xml file
as well as any number of binary files (raw numeric buffers) which contain
information about their start time in the file name. An example directory
might look like this:

data_folder
   metadata.xml
   DAS_20240530T011500_000000Z.raw
   DAS_20240530T011501_000000Z.raw
"""
from __future__ import annotations
from .core import XMLBinaryV1
